"""
Module which will convert from
superannotate annotation format to other annotation formats
"""
import copy
import json
import shutil
from pathlib import Path

import numpy as np
from superannotate.logger import get_default_logger

from .converters.converters import Converter

logger = get_default_logger()


def _load_files(path_to_imgs, task, ptype):
    suffix = None
    if ptype == "Pixel":
        suffix = "___pixel.json"
    else:
        suffix = "___objects.json"

    orig_images = path_to_imgs.glob("*" + suffix)
    orig_images = [str(x).replace(suffix, "") for x in orig_images]
    all_files = None

    if task == "keypoint_detection":
        all_files = np.array([[fname, fname + suffix] for fname in orig_images])
    elif ptype == "Pixel":
        all_files = np.array(
            [[fname, fname + suffix, fname + "___save.png"] for fname in orig_images]
        )
    elif ptype == "Vector":
        all_files = np.array(
            [[fname, fname + suffix, fname + "___fuse.png"] for fname in orig_images]
        )

    return all_files


def _move_files(data_set, src):
    train_path = src / "image_set"
    if data_set is not None:
        for tup in data_set:
            for i in tup:
                if Path(i).exists():
                    shutil.copy(i, train_path / Path(i).name)
    else:
        logger.warning("Images doesn't exist")


def _create_classes_mapper(imgs, classes_json):
    classes = {}
    j_data = json.load(open(classes_json))
    for instance in j_data:
        classes[instance["name"]] = instance["id"]

    with open(imgs / "image_set" / "classes_mapper.json", "w") as fp:
        json.dump(classes, fp)


def export_from_sa(args):
    """
    :param args: All arguments that will be used during convertion.
    :type args: Namespace
    """
    data_set = None

    (args.output_dir / "image_set").mkdir(parents=True, exist_ok=True)

    try:
        _create_classes_mapper(
            args.output_dir, args.input_dir / "classes" / "classes.json"
        )
    except Exception as e:
        logger.debug(str(e), exc_info=True)
        _create_classes_mapper(args.input_dir, args.output_dir)

    data_set = _load_files(args.input_dir, args.task, args.project_type)
    _move_files(data_set, args.output_dir)

    args.__dict__.update(
        {"direction": "to", "export_root": args.output_dir / "image_set"}
    )
    converter = Converter(args)

    if data_set is not None:
        converter.strategy.set_dataset_name(args.dataset_name)
        converter.convert_from_sa()

    image_set_failed = copy.deepcopy(converter.strategy.failed_conversion_cnt)

    logger.info("Conversion completed")
    return image_set_failed



def _create_attribute_list(attribute_dict, class_name, class_id_map):
    attr_keys = attribute_dict.keys()
    attributes = []
    for attr in attr_keys:
        if attr in class_id_map[class_name]["attribute_groups"].keys():
            group_name = attr
            if isinstance(attribute_dict[attr], dict):
                keys = attribute_dict[attr].keys()
            else:
                keys = [attribute_dict[attr]]

            for key in keys:
                if (
                    key
                    not in class_id_map[class_name]["attribute_groups"][attr][
                        "attributes"
                    ]
                ):
                    class_id_map[class_name]["attribute_groups"][attr][
                        "attributes"
                    ].append(key)
                dd = {"name": key, "groupName": group_name}
                attributes.append(dd.copy())

        else:
            multi = 0
            if isinstance(attribute_dict[attr], dict):
                multi = 1

            class_id_map[class_name]["attribute_groups"][attr] = {
                "is_multiselect": multi,
                "attributes": [],
            }
    return attributes



class AppException(Exception):
    """
    Base exception for Licensee App. All exceptions thrown by inviter should
    extend this.
    """

    def __init__(self, message):
        super().__init__(message)

        self.message = str(message)

    def __str__(self):
        return self.message


class BackendError(AppException):
    """
    Backend Error
    """


class AppValidationException(AppException):
    """
    App validation exception
    """


class ImageProcessingException(AppException):
    """
    App validation exception
    """



import os
import sys
import tempfile
from typing import Any
from typing import Optional

import lib.core as constances
from lib import __file__ as lib_path
from lib.app.helpers import split_project_path
from lib.app.input_converters.conversion import import_annotation
from lib.app.interface.base_interface import BaseInterfaceFacade
from lib.app.interface.sdk_interface import attach_document_urls_to_project
from lib.app.interface.sdk_interface import attach_image_urls_to_project
from lib.app.interface.sdk_interface import attach_video_urls_to_project
from lib.app.interface.sdk_interface import create_folder
from lib.app.interface.sdk_interface import create_project
from lib.app.interface.sdk_interface import upload_annotations_from_folder_to_project
from lib.app.interface.sdk_interface import upload_images_from_folder_to_project
from lib.app.interface.sdk_interface import upload_preannotations_from_folder_to_project
from lib.app.interface.sdk_interface import upload_videos_from_folder_to_project
from lib.core.entities import ConfigEntity
from lib.infrastructure.controller import Controller
from lib.infrastructure.repositories import ConfigRepository


class CLIFacade(BaseInterfaceFacade):
    """
    With SuperAnnotate CLI, basic tasks can be accomplished using shell commands:
    superannotatecli <command> <--arg1 val1> <--arg2 val2> [--optional_arg3 val3] [--optional_arg4] ...
    """

    @staticmethod
    def version():
        """
        To show the version of the current SDK installation
        """
        with open(
            f"{os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(lib_path))))}/version.py"
        ) as f:
            version = f.read().rstrip()[15:-1]
            print(version)
        sys.exit(0)

    @staticmethod
    def init():
        """
        To initialize CLI (and SDK) with team token
        """
        repo = ConfigRepository()
        config = repo.get_one(uuid=constances.TOKEN_UUID)
        if config:
            if not input(
                f"File {repo.config_path} exists. Do you want to overwrite? [y/n] : "
            ).lower() in ("y", "yes"):
                return
        token = input(
            "Input the team SDK token from https://app.superannotate.com/team : "
        )
        config_entity = ConfigEntity(uuid=constances.TOKEN_UUID, value=token)
        repo.insert(config_entity)
        if config:
            print("Configuration file successfully updated.")
        else:
            print("Configuration file successfully created.")
        sys.exit(0)

    def create_project(self, name: str, description: str, type: str):
        """
        To create a new project
        """
        create_project(name, description, type)

    def create_folder(self, project: str, name: str):
        """
        To create a new folder
        """
        create_folder(project, name)
        sys.exit(0)

    def upload_images(
        self,
        project: str,
        folder: str,
        extensions: str = constances.DEFAULT_IMAGE_EXTENSIONS,
        set_annotation_status: str = constances.AnnotationStatus.NOT_STARTED.name,
        exclude_file_patterns=constances.DEFAULT_FILE_EXCLUDE_PATTERNS,
        recursive_subfolders=False,
        image_quality_in_editor=None,
    ):
        """
        To upload images from folder to project use:

        If optional argument recursive is given then subfolders of <folder_path> are also
         recursively scanned for available images.
        Optional argument extensions accepts comma separated list of image extensions to look for.
        If the argument is not given then value jpg,jpeg,png,tif,tiff,webp,bmp is assumed.
        """
        if not isinstance(extensions, list):
            extensions = extensions.split(",")
        upload_images_from_folder_to_project(
            project,
            folder_path=folder,
            extensions=extensions,
            annotation_status=set_annotation_status,
            exclude_file_patterns=exclude_file_patterns,
            recursive_subfolders=recursive_subfolders,
            image_quality_in_editor=image_quality_in_editor,
        )
        sys.exit(0)

    def export_project(
        self,
        project,
        folder,
        include_fuse=False,
        disable_extract_zip_contents=False,
        annotation_statuses=None,
    ):
        project_name, folder_name = split_project_path(project)
        folders = None
        if folder_name:
            folders = [folder_name]
        export_res = Controller.get_default().prepare_export(
            project_name, folders, include_fuse, False, annotation_statuses
        )
        export_name = export_res.data["name"]

        use_case = Controller.get_default().download_export(
            project_name=project_name,
            export_name=export_name,
            folder_path=folder,
            extract_zip_contents=not disable_extract_zip_contents,
            to_s3_bucket=False,
        )
        if use_case.is_valid():
            for _ in use_case.execute():
                continue

        sys.exit(0)

    def upload_preannotations(
        self, project, folder, dataset_name=None, task=None, format=None
    ):
        """
        To upload preannotations from folder to project use
        Optional argument format accepts input annotation format. It can have COCO or SuperAnnotate values.
         If the argument is not given then SuperAnnotate (the native annotation format) is assumed.
        Only when COCO format is specified dataset-name and task arguments are required.
        dataset-name specifies JSON filename (without extension) in <folder_path>.
        task specifies the COCO task for conversion. Please see import_annotation_format for more details.
        """
        self._upload_annotations(
            project=project,
            folder=folder,
            format=format,
            dataset_name=dataset_name,
            task=task,
            pre=True,
        )
        sys.exit(0)

    def upload_annotations(
        self, project, folder, dataset_name=None, task=None, format=None
    ):
        """
        To upload annotations from folder to project use
        Optional argument format accepts input annotation format. It can have COCO or SuperAnnotate values.
        If the argument is not given then SuperAnnotate (the native annotation format) is assumed.
        Only when COCO format is specified dataset-name and task arguments are required.
        dataset-name specifies JSON filename (without extension) in <folder_path>.
        task specifies the COCO task for conversion. Please see import_annotation_format for more details.
        """
        self._upload_annotations(
            project=project,
            folder=folder,
            format=format,
            dataset_name=dataset_name,
            task=task,
            pre=False,
        )
        sys.exit(0)

    def _upload_annotations(
        self, project, folder, format, dataset_name, task, pre=True
    ):
        project_folder_name = project
        project_name, folder_name = split_project_path(project)
        project = (
            Controller.get_default()
            .get_project_metadata(project_name=project_name)
            .data
        )
        if not format:
            format = "SuperAnnotate"
        if not dataset_name and format == "COCO":
            raise Exception("Data-set name is required")
        elif not dataset_name:
            dataset_name = ""
        if not task:
            task = "object_detection"
        annotations_path = folder
        with tempfile.TemporaryDirectory() as temp_dir:
            if format != "SuperAnnotate":
                import_annotation(
                    input_dir=folder,
                    output_dir=temp_dir,
                    dataset_format=format,
                    dataset_name=dataset_name,
                    project_type=constances.ProjectType.get_name(
                        project["project"].type
                    ),
                    task=task,
                )
                annotations_path = temp_dir
            if pre:
                upload_preannotations_from_folder_to_project(
                    project_folder_name, annotations_path
                )
            else:
                upload_annotations_from_folder_to_project(
                    project_folder_name, annotations_path
                )
        sys.exit(0)

    def attach_image_urls(
        self, project: str, attachments: str, annotation_status: Optional[Any] = None
    ):
        """
        To attach image URLs to project use:
        """

        attach_image_urls_to_project(
            project=project,
            attachments=attachments,
            annotation_status=annotation_status,
        )
        sys.exit(0)

    def attach_video_urls(
        self, project: str, attachments: str, annotation_status: Optional[Any] = None
    ):
        attach_video_urls_to_project(
            project=project,
            attachments=attachments,
            annotation_status=annotation_status,
        )
        sys.exit(0)

    @staticmethod
    def attach_document_urls(
        project: str, attachments: str, annotation_status: Optional[Any] = None
    ):
        attach_document_urls_to_project(
            project=project,
            attachments=attachments,
            annotation_status=annotation_status,
        )
        sys.exit(0)

    def upload_videos(
        self,
        project,
        folder,
        target_fps=None,
        recursive=False,
        extensions=constances.DEFAULT_VIDEO_EXTENSIONS,
        set_annotation_status=constances.AnnotationStatus.NOT_STARTED.name,
        start_time=0.0,
        end_time=None,
    ):
        """
        To upload videos from folder to project use
        If optional argument recursive is given then subfolders of <folder_path> are also recursively scanned for available videos.
        Optional argument extensions accepts comma separated list of image extensions to look for.
        If the argument is not given then value mp4,avi,mov,webm,flv,mpg,ogg is assumed.
        target-fps specifies how many frames per second need to extract from the videos (approximate).
        If not specified all frames will be uploaded.
        start-time specifies time (in seconds) from which to start extracting frames, default is 0.0.
        end-time specifies time (in seconds) up to which to extract frames. If it is not specified, then up to end is assumed.
        """

        upload_videos_from_folder_to_project(
            project=project,
            folder_path=folder,
            extensions=extensions,
            exclude_file_patterns=(),
            recursive_subfolders=recursive,
            target_fps=target_fps,
            start_time=start_time,
            end_time=end_time,
            annotation_status=set_annotation_status,
            image_quality_in_editor=None,
        )
        sys.exit(0)



import concurrent.futures
import io
import json
import os
from collections import namedtuple
from typing import List
from typing import Optional
from typing import Tuple

import boto3
import lib.core as constances
from lib.core.conditions import Condition
from lib.core.conditions import CONDITION_EQ as EQ
from lib.core.data_handlers import ChainedAnnotationHandlers
from lib.core.data_handlers import DocumentTagHandler
from lib.core.data_handlers import LastActionHandler
from lib.core.data_handlers import MissingIDsHandler
from lib.core.data_handlers import VideoFormatHandler
from lib.core.entities import AnnotationClassEntity
from lib.core.entities import FolderEntity
from lib.core.entities import ImageEntity
from lib.core.entities import ProjectEntity
from lib.core.entities import TeamEntity
from lib.core.exceptions import AppException
from lib.core.reporter import Reporter
from lib.core.repositories import BaseManageableRepository
from lib.core.service_types import UploadAnnotationAuthData
from lib.core.serviceproviders import SuperannotateServiceProvider
from lib.core.types import PriorityScore
from lib.core.usecases.base import BaseReportableUseCae
from lib.core.usecases.images import GetBulkImages
from lib.core.usecases.images import ValidateAnnotationUseCase
from lib.core.video_convertor import VideoFrameGenerator
from superannotate.logger import get_default_logger
from superannotate_schemas.validators import AnnotationValidators

logger = get_default_logger()


class UploadAnnotationsUseCase(BaseReportableUseCae):
    MAX_WORKERS = 10
    CHUNK_SIZE = 100
    AUTH_DATA_CHUNK_SIZE = 500
    ImageInfo = namedtuple("ImageInfo", ["path", "name", "id"])

    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        team: TeamEntity,
        images: BaseManageableRepository,
        annotation_classes: List[AnnotationClassEntity],
        annotation_paths: List[str],
        backend_service_provider: SuperannotateServiceProvider,
        templates: List[dict],
        validators: AnnotationValidators,
        pre_annotation: bool = False,
        client_s3_bucket=None,
        folder_path: str = None,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._team = team
        self._images = images
        self._backend_service = backend_service_provider
        self._annotation_classes = annotation_classes
        self._annotation_paths = annotation_paths
        self._client_s3_bucket = client_s3_bucket
        self._pre_annotation = pre_annotation
        self._templates = templates
        self._annotations_to_upload = []
        self._missing_annotations = []
        self._validators = validators
        self.missing_attribute_groups = set()
        self.missing_classes = set()
        self.missing_attributes = set()
        self._folder_path = folder_path

    @property
    def annotation_postfix(self):
        if self._project.type in (
            constances.ProjectType.VIDEO.value,
            constances.ProjectType.DOCUMENT.value,
        ):
            return constances.ATTACHED_VIDEO_ANNOTATION_POSTFIX
        elif self._project.type == constances.ProjectType.VECTOR.value:
            return constances.VECTOR_ANNOTATION_POSTFIX
        elif self._project.type == constances.ProjectType.PIXEL.value:
            return constances.PIXEL_ANNOTATION_POSTFIX

    @staticmethod
    def extract_name(value: str):
        return os.path.basename(
            value.replace(constances.PIXEL_ANNOTATION_POSTFIX, "")
            .replace(constances.VECTOR_ANNOTATION_POSTFIX, "")
            .replace(constances.ATTACHED_VIDEO_ANNOTATION_POSTFIX, ""),
        )

    @property
    def annotations_to_upload(self):
        if not self._annotations_to_upload:
            annotation_paths = self._annotation_paths
            images_detail = []
            for annotation_path in annotation_paths:
                images_detail.append(
                    self.ImageInfo(
                        id=None,
                        path=annotation_path,
                        name=self.extract_name(annotation_path),
                    )
                )
            images_data = (
                GetBulkImages(
                    service=self._backend_service,
                    project_id=self._project.id,
                    team_id=self._project.team_id,
                    folder_id=self._folder.uuid,
                    images=[image.name for image in images_detail],
                )
                .execute()
                .data
            )
            for image_data in images_data:
                for idx, detail in enumerate(images_detail):
                    if detail.name == image_data.name:
                        images_detail[idx] = detail._replace(id=image_data.uuid)
                        break

            missing_annotations = list(
                filter(lambda image_detail: image_detail.id is None, images_detail)
            )
            annotations_to_upload = list(
                filter(lambda image_detail: image_detail.id is not None, images_detail)
            )
            if missing_annotations:
                logger.warning(
                    f"Couldn't find {len(missing_annotations)}/{len(annotations_to_upload + missing_annotations)} "
                    "items on the platform that match the annotations you want to upload."
                )
            self._missing_annotations = missing_annotations
            self._annotations_to_upload = annotations_to_upload
        return self._annotations_to_upload

    @property
    def missing_annotations(self):
        if not self._missing_annotations:
            self._missing_annotations = []
        return self._missing_annotations

    def get_annotation_upload_data(
        self, image_ids: List[int]
    ) -> UploadAnnotationAuthData:
        if self._pre_annotation:
            function = self._backend_service.get_pre_annotation_upload_data
        else:
            function = self._backend_service.get_annotation_upload_data
        response = function(
            project_id=self._project.id,
            team_id=self._project.team_id,
            folder_id=self._folder.uuid,
            image_ids=image_ids,
        )
        if response.ok:
            return response.data

    def _upload_annotation(
        self,
        image_id: int,
        image_name: str,
        upload_data: UploadAnnotationAuthData,
        path: str,
        bucket,
    ):
        try:
            self.reporter.disable_warnings()
            response = UploadAnnotationUseCase(
                project=self._project,
                folder=self._folder,
                team=self._team,
                image=ImageEntity(
                    uuid=image_id,
                    name=image_name,
                    team_id=self._project.team_id,
                    project_id=self._project.id,
                ),
                images=self._images,
                annotation_classes=self._annotation_classes,
                backend_service_provider=self._backend_service,
                reporter=self.reporter,
                templates=self._templates,
                annotation_upload_data=upload_data,
                client_s3_bucket=self._client_s3_bucket,
                annotation_path=path,
                verbose=False,
                s3_bucket=bucket,
                validators=self._validators,
            ).execute()
            if response.errors:
                return path, False
            return path, True
        except Exception as e:
            logger.debug(str(e), exc_info=True)
            return path, False
        finally:
            self.reporter.enable_warnings()

    def get_bucket_to_upload(self, ids: List[int]):
        upload_data = self.get_annotation_upload_data(ids)
        if upload_data:
            session = boto3.Session(
                aws_access_key_id=upload_data.access_key,
                aws_secret_access_key=upload_data.secret_key,
                aws_session_token=upload_data.session_token,
                region_name=upload_data.region,
            )
            resource = session.resource("s3")
            return resource.Bucket(upload_data.bucket)

    def _log_report(self):
        for key, values in self.reporter.custom_messages.items():
            if key in [
                "missing_classes",
                "missing_attribute_groups",
                "missing_attributes",
            ]:
                template = key + ": {}"
                if key == "missing_classes":
                    template = "Could not find annotation classes matching existing classes on the platform: [{}]"
                elif key == "missing_attribute_groups":
                    template = (
                        "Could not find attribute groups matching existing attribute groups"
                        " on the platform: [{}]"
                    )
                elif key == "missing_attributes":
                    template = "Could not find attributes matching existing attributes on the platform: [{}]"
                logger.warning(template.format("', '".join(values)))
        if self.reporter.custom_messages.get("invalid_jsons"):
            logger.warning(
                f"Couldn't validate {len(self.reporter.custom_messages['invalid_jsons'])}/"
                f"{len(self.annotations_to_upload + self.missing_annotations)} annotations from {self._folder_path}. "
                f"{constances.USE_VALIDATE_MESSAGE}"
            )

    def execute(self):
        uploaded_annotations = []
        failed_annotations = []
        if self.annotations_to_upload:
            iterations_range = range(
                0, len(self.annotations_to_upload), self.AUTH_DATA_CHUNK_SIZE
            )
            self.reporter.start_progress(
                len(self.annotations_to_upload), description="Uploading Annotations"
            )
            for step in iterations_range:
                annotations_to_upload = self.annotations_to_upload[
                    step : step + self.AUTH_DATA_CHUNK_SIZE  # noqa: E203
                ]
                upload_data = self.get_annotation_upload_data(
                    [int(image.id) for image in annotations_to_upload]
                )
                bucket = self.get_bucket_to_upload(
                    [int(image.id) for image in annotations_to_upload]
                )
                if bucket:
                    image_id_name_map = {
                        image.id: image for image in self.annotations_to_upload
                    }
                    # dummy progress
                    for _ in range(
                        len(annotations_to_upload) - len(upload_data.images)
                    ):
                        self.reporter.update_progress()
                    with concurrent.futures.ThreadPoolExecutor(
                        max_workers=self.MAX_WORKERS
                    ) as executor:
                        results = [
                            executor.submit(
                                self._upload_annotation,
                                image_id,
                                image_id_name_map[image_id].name,
                                upload_data,
                                image_id_name_map[image_id].path,
                                bucket,
                            )
                            for image_id, image_data in upload_data.images.items()
                        ]
                        for future in concurrent.futures.as_completed(results):
                            annotation, uploaded = future.result()
                            if uploaded:
                                uploaded_annotations.append(annotation)
                            else:
                                failed_annotations.append(annotation)
                            self.reporter.update_progress()
            self.reporter.finish_progress()
            self._log_report()
        self._response.data = (
            uploaded_annotations,
            failed_annotations,
            [annotation.path for annotation in self.missing_annotations],
        )
        return self._response


class UploadAnnotationUseCase(BaseReportableUseCae):
    def __init__(
        self,
        project: ProjectEntity,
        folder: FolderEntity,
        image: ImageEntity,
        images: BaseManageableRepository,
        team: TeamEntity,
        annotation_classes: List[AnnotationClassEntity],
        backend_service_provider: SuperannotateServiceProvider,
        reporter: Reporter,
        templates: List[dict],
        validators: AnnotationValidators,
        annotation_upload_data: UploadAnnotationAuthData = None,
        annotations: dict = None,
        s3_bucket=None,
        client_s3_bucket=None,
        mask=None,
        verbose: bool = True,
        annotation_path: str = None,
        pass_validation: bool = False,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._image = image
        self._images = images
        self._team = team
        self._backend_service = backend_service_provider
        self._annotation_classes = annotation_classes
        self._annotation_json = annotations
        self._mask = mask
        self._verbose = verbose
        self._templates = templates
        self._annotation_path = annotation_path
        self._annotation_upload_data = annotation_upload_data
        self._s3_bucket = s3_bucket
        self._client_s3_bucket = client_s3_bucket
        self._pass_validation = pass_validation
        self._validators = validators

    @property
    def annotation_upload_data(self) -> UploadAnnotationAuthData:
        if not self._annotation_upload_data:
            response = self._backend_service.get_annotation_upload_data(
                project_id=self._project.id,
                team_id=self._project.team_id,
                folder_id=self._folder.uuid,
                image_ids=[self._image.uuid],
            )
            if response.ok:
                self._annotation_upload_data = response.data
        return self._annotation_upload_data

    @property
    def s3_bucket(self):
        if not self._s3_bucket:
            upload_data = self.annotation_upload_data
            if upload_data:
                session = boto3.Session(
                    aws_access_key_id=upload_data.access_key,
                    aws_secret_access_key=upload_data.secret_key,
                    aws_session_token=upload_data.session_token,
                    region_name=upload_data.region,
                )
                resource = session.resource("s3")
                self._s3_bucket = resource.Bucket(upload_data.bucket)
        return self._s3_bucket

    def get_s3_file(self, s3, path: str):
        file = io.BytesIO()
        s3_object = s3.Object(self._client_s3_bucket, path)
        s3_object.download_fileobj(file)
        file.seek(0)
        return file

    @property
    def from_s3(self):
        if self._client_s3_bucket:
            from_session = boto3.Session()
            return from_session.resource("s3")

    def set_annotation_json(self):
        if not self._annotation_json:
            if self._client_s3_bucket:
                self._annotation_json = json.load(
                    self.get_s3_file(self.from_s3, self._annotation_path)
                )
                if self._project.type == constances.ProjectType.PIXEL.value:
                    self._mask = self.get_s3_file(
                        self.from_s3,
                        self._annotation_path.replace(
                            constances.PIXEL_ANNOTATION_POSTFIX,
                            constances.ANNOTATION_MASK_POSTFIX,
                        ),
                    )
            else:
                self._annotation_json = json.load(open(self._annotation_path))
                if self._project.type == constances.ProjectType.PIXEL.value:
                    self._mask = open(
                        self._annotation_path.replace(
                            constances.PIXEL_ANNOTATION_POSTFIX,
                            constances.ANNOTATION_MASK_POSTFIX,
                        ),
                        "rb",
                    )

    @staticmethod
    def prepare_annotations(
        project_type: int,
        annotations: dict,
        annotation_classes: List[AnnotationClassEntity],
        templates: List[dict],
        reporter: Reporter,
        team: TeamEntity,
    ) -> dict:
        handlers_chain = ChainedAnnotationHandlers()
        if project_type in (
            constances.ProjectType.VECTOR.value,
            constances.ProjectType.PIXEL.value,
            constances.ProjectType.DOCUMENT.value,
        ):
            handlers_chain.attach(
                MissingIDsHandler(annotation_classes, templates, reporter)
            )
        elif project_type == constances.ProjectType.VIDEO.value:
            handlers_chain.attach(VideoFormatHandler(annotation_classes, reporter))
        if project_type == constances.ProjectType.DOCUMENT.value:
            handlers_chain.attach(DocumentTagHandler(annotation_classes))
        handlers_chain.attach(LastActionHandler(team.creator_id))
        return handlers_chain.handle(annotations)

    def clean_json(self, json_data: dict,) -> Tuple[bool, dict]:
        use_case = ValidateAnnotationUseCase(
            constances.ProjectType.get_name(self._project.type),
            annotation=json_data,
            validators=self._validators,
        )
        return use_case.execute().data

    def execute(self):
        if self.is_valid():
            self.set_annotation_json()
            is_valid, clean_json = self.clean_json(self._annotation_json)
            if is_valid:
                self._annotation_json = clean_json
                bucket = self.s3_bucket
                annotation_json = self.prepare_annotations(
                    project_type=self._project.type,
                    annotations=self._annotation_json,
                    annotation_classes=self._annotation_classes,
                    templates=self._templates,
                    reporter=self.reporter,
                    team=self._team,
                )
                bucket.put_object(
                    Key=self.annotation_upload_data.images[self._image.uuid][
                        "annotation_json_path"
                    ],
                    Body=json.dumps(annotation_json),
                )
                if (
                    self._project.type == constances.ProjectType.PIXEL.value
                    and self._mask
                ):
                    bucket.put_object(
                        Key=self.annotation_upload_data.images[self._image.uuid][
                            "annotation_bluemap_path"
                        ],
                        Body=self._mask,
                    )
                self._image.annotation_status_code = (
                    constances.AnnotationStatus.IN_PROGRESS.value
                )
                self._images.update(self._image)
                if self._verbose:
                    self.reporter.log_info(
                        f"Uploading annotations for image {str(self._image.name)} in project {self._project.name}."
                    )
            else:
                self._response.errors = constances.INVALID_JSON_MESSAGE
                self.reporter.store_message("invalid_jsons", self._annotation_path)
                self.reporter.log_warning(
                    f"Couldn't validate annotations. {constances.USE_VALIDATE_MESSAGE}"
                )
        return self._response


class GetAnnotations(BaseReportableUseCae):
    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        images: BaseManageableRepository,
        item_names: Optional[List[str]],
        backend_service_provider: SuperannotateServiceProvider,
        show_process: bool = True,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._images = images
        self._item_names = item_names
        self._client = backend_service_provider
        self._show_process = show_process
        self._item_names_provided = True

    def validate_project_type(self):
        if self._project.type == constances.ProjectType.PIXEL.value:
            raise AppException("The function is not supported for Pixel projects.")

    def validate_item_names(self):
        if self._item_names:
            item_names = list(dict.fromkeys(self._item_names))
            len_unique_items, len_items = len(item_names), len(self._item_names)
            if len_unique_items < len_items:
                self.reporter.log_info(
                    f"Dropping duplicates. Found {len_unique_items}/{len_items} unique items."
                )
                self._item_names = item_names
        else:
            self._item_names_provided = False
            condition = (
                Condition("team_id", self._project.team_id, EQ)
                & Condition("project_id", self._project.id, EQ)
                & Condition("folder_id", self._folder.uuid, EQ)
            )

            self._item_names = [item.name for item in self._images.get_all(condition)]

    def _prettify_annotations(self, annotations: List[dict]):
        if self._item_names_provided:
            try:
                data = []
                for annotation in annotations:
                    data.append(
                        (
                            self._item_names.index(annotation["metadata"]["name"]),
                            annotation,
                        )
                    )
                return [i[1] for i in sorted(data, key=lambda x: x[0])]
            except KeyError:
                raise AppException("Broken data.")
        return annotations

    def execute(self):
        if self.is_valid():
            items_count = len(self._item_names)
            self.reporter.log_info(
                f"Getting {items_count} annotations from "
                f"{self._project.name}{f'/{self._folder.name}' if self._folder.name != 'root' else ''}."
            )
            self.reporter.start_progress(items_count, disable=not self._show_process)
            annotations = self._client.get_annotations(
                team_id=self._project.team_id,
                project_id=self._project.id,
                folder_id=self._folder.uuid,
                items=self._item_names,
                reporter=self.reporter,
            )
            received_items_count = len(annotations)
            self.reporter.finish_progress()
            if items_count > received_items_count:
                self.reporter.log_warning(
                    f"Could not find annotations for {items_count - received_items_count}/{items_count} items."
                )
            self._response.data = self._prettify_annotations(annotations)
        return self._response


class GetVideoAnnotationsPerFrame(BaseReportableUseCae):
    def __init__(
        self,
        reporter: Reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        images: BaseManageableRepository,
        video_name: str,
        fps: int,
        backend_service_provider: SuperannotateServiceProvider,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._images = images
        self._video_name = video_name
        self._fps = fps
        self._client = backend_service_provider

    def validate_project_type(self):
        if self._project.type != constances.ProjectType.VIDEO.value:
            raise AppException(
                "The function is not supported for"
                f" {constances.ProjectType.get_name(self._project.type)} projects."
            )

    def execute(self):
        if self.is_valid():
            self.reporter.disable_info()
            response = GetAnnotations(
                reporter=self.reporter,
                project=self._project,
                folder=self._folder,
                images=self._images,
                item_names=[self._video_name],
                backend_service_provider=self._client,
                show_process=False,
            ).execute()
            self.reporter.enable_info()
            if response.data:
                generator = VideoFrameGenerator(response.data[0], fps=self._fps)

                self.reporter.log_info(
                    f"Getting annotations for {generator.frames_count} frames from {self._video_name}."
                )
                if response.errors:
                    self._response.errors = response.errors
                    return self._response
                if not response.data:
                    self._response.errors = AppException(
                        f"Video {self._video_name} not found."
                    )
                annotations = response.data
                if annotations:
                    self._response.data = list(generator)
                else:
                    self._response.data = []
            else:
                self._response.errors = "Couldn't get annotations."
        return self._response


class UploadPriorityScoresUseCase(BaseReportableUseCae):
    CHUNK_SIZE = 100

    def __init__(
        self,
        reporter,
        project: ProjectEntity,
        folder: FolderEntity,
        scores: List[PriorityScore],
        project_folder_name: str,
        backend_service_provider: SuperannotateServiceProvider,
    ):
        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._scores = scores
        self._client = backend_service_provider
        self._project_folder_name = project_folder_name

    @staticmethod
    def get_clean_priority(priority):
        if len(str(priority)) > 8:
            priority = float(str(priority)[:8])
        if priority > 1000000:
            priority = 1000000
        if priority < 0:
            priority = 0
        if str(float(priority)).split(".")[1:2]:
            if len(str(float(priority)).split(".")[1]) > 5:
                priority = float(
                    str(float(priority)).split(".")[0]
                    + "."
                    + str(float(priority)).split(".")[1][:5]
                )
        return priority

    @property
    def folder_path(self):
        return f"{self._project.name}{f'/{self._folder.name}' if self._folder.name != 'root' else ''}"

    @property
    def uploading_info(self):
        data_len: int = len(self._scores)
        return (
            f"Uploading  priority scores for {data_len} item(s) to {self.folder_path}."
        )

    def execute(self):
        if self.is_valid():
            priorities = []
            initial_scores = []
            for i in self._scores:
                priorities.append(
                    {
                        "name": i.name,
                        "entropy_value": self.get_clean_priority(i.priority),
                    }
                )
                initial_scores.append(i.name)
            uploaded_score_names = []
            self.reporter.log_info(self.uploading_info)
            iterations = range(0, len(priorities), self.CHUNK_SIZE)
            self.reporter.start_progress(iterations, "Uploading priority scores")
            if iterations:
                for i in iterations:
                    priorities_to_upload = priorities[
                        i : i + self.CHUNK_SIZE
                    ]  # noqa: E203
                    res = self._client.upload_priority_scores(
                        team_id=self._project.team_id,
                        project_id=self._project.id,
                        folder_id=self._folder.uuid,
                        priorities=priorities_to_upload,
                    )
                    self.reporter.update_progress(len(priorities_to_upload))
                    uploaded_score_names.extend(
                        list(map(lambda x: x["name"], res.get("data", [])))
                    )
                self.reporter.finish_progress()
                skipped_score_names = list(
                    set(initial_scores) - set(uploaded_score_names)
                )
                self._response.data = (uploaded_score_names, skipped_score_names)
            else:
                self.reporter.warning_messages("Empty scores.")
        return self._response



import json
import os
import sys
import time

import numpy as np
from PIL import Image
from superannotate.logger import get_default_logger
from tqdm import tqdm

logger = get_default_logger()

_PROJECT_TYPES = {"Vector": 1, "Pixel": 2}

_ANNOTATION_STATUSES = {
    "NotStarted": 1,
    "InProgress": 2,
    "QualityCheck": 3,
    "Returned": 4,
    "Completed": 5,
    "Skipped": 6,
}


def hex_to_rgb(hex_string):
    """Converts HEX values to RGB values
    """
    h = hex_string.lstrip("#")
    return tuple(int(h[i : i + 2], 16) for i in (0, 2, 4))


def blue_color_generator(n, hex_values=True):
    """ Blue colors generator for SuperAnnotate blue mask.
    """
    hex_colors = []
    for i in range(n + 1):
        int_color = i * 15
        bgr_color = np.array(
            [int_color & 255, (int_color >> 8) & 255, (int_color >> 16) & 255, 255],
            dtype=np.uint8,
        )
        hex_color = (
            "#"
            + "{:02x}".format(bgr_color[2])
            + "{:02x}".format(bgr_color[1],)
            + "{:02x}".format(bgr_color[0])
        )
        if hex_values:
            hex_colors.append(hex_color)
        else:
            hex_colors.append(hex_to_rgb(hex_color))
    return hex_colors[1:]


def id2rgb(id_map):
    if isinstance(id_map, np.ndarray):
        id_map_copy = id_map.copy()
        rgb_shape = tuple(list(id_map.shape) + [3])
        rgb_map = np.zeros(rgb_shape, dtype=np.uint8)
        for i in range(3):
            rgb_map[..., i] = id_map_copy % 256
            id_map_copy //= 256
        return rgb_map
    color = []
    for _ in range(3):
        color.append(id_map % 256)
        id_map //= 256
    return color


def save_desktop_format(output_dir, classes, files_dict):
    cat_id_map = {}
    new_classes = []
    for idx, class_ in enumerate(classes):
        cat_id_map[class_["name"]] = idx + 2
        class_["id"] = idx + 2
        new_classes.append(class_)
    with open(output_dir.joinpath("classes.json"), "w") as fw:
        json.dump(new_classes, fw)

    meta = {
        "type": "meta",
        "name": "lastAction",
        "timestamp": int(round(time.time() * 1000)),
    }
    new_json = {}
    files_path = []
    (output_dir / "images" / "thumb").mkdir()
    for file_name, json_data in files_dict.items():
        file_name = file_name.replace("___objects.json", "")
        if not (output_dir / "images" / file_name).exists():
            continue

        for js_data in json_data:
            if "className" in js_data:
                js_data["classId"] = cat_id_map[js_data["className"]]
        json_data.append(meta)
        new_json[file_name] = json_data

        files_path.append(
            {
                "srcPath": str(output_dir.resolve() / file_name),
                "name": file_name,
                "imagePath": str(output_dir.resolve() / file_name),
                "thumbPath": str(
                    output_dir.resolve()
                    / "images"
                    / "thumb"
                    / ("thmb_" + file_name + ".jpg")
                ),
                "valid": True,
            }
        )

        img = Image.open(output_dir / "images" / file_name)
        img.thumbnail((168, 120), Image.ANTIALIAS)
        img.save(output_dir / "images" / "thumb" / ("thmb_" + file_name + ".jpg"))

    with open(output_dir / "images" / "images.sa", "w") as fw:
        fw.write(json.dumps(files_path))

    with open(output_dir.joinpath("annotations.json"), "w") as fw:
        json.dump(new_json, fw)

    with open(output_dir / "config.json", "w") as fw:
        json.dump({"pathSeparator": os.sep, "os": sys.platform}, fw)


def save_web_format(output_dir, classes, files_dict):
    for key, value in files_dict.items():
        with open(output_dir.joinpath(key), "w") as fw:
            json.dump(value, fw, indent=2)

    with open(output_dir.joinpath("classes", "classes.json"), "w") as fw:
        json.dump(classes, fw)


def write_to_json(output_path, json_data):
    with open(output_path, "w") as fw:
        json.dump(json_data, fw, indent=2)


MAX_IMAGE_SIZE = 100 * 1024 * 1024  # 100 MB limit


def tqdm_converter(total_num, images_converted, images_not_converted, finish_event):
    with tqdm(total=total_num) as pbar:
        while True:
            finished = finish_event.wait(5)
            if not finished:
                sum_all = len(images_converted) + len(images_not_converted)
                pbar.update(sum_all - pbar.n)
            else:
                pbar.update(total_num - pbar.n)
                break


def get_annotation_json_name(image_name, project_type):
    if project_type == "Vector":
        return image_name + "___objects.json"
    else:
        return image_name + "___pixel.json"


def get_annotation_png_name(image_name):
    return image_name + "___save.png"



from pathlib import Path

import numpy as np

from ....common import write_to_json
from ..baseStrategy import baseStrategy


class GoogleCloudStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        path = Path(self.export_root).joinpath(self.dataset_name + ".csv")
        classes = self.conversion_algorithm(path, self.output_dir)
        sa_classes = self._create_classes(classes)
        (self.output_dir / "classes").mkdir(exist_ok=True)
        write_to_json(self.output_dir / "classes" / "classes.json", sa_classes)

    def _create_classes(self, classes):
        classes_loader = []
        for class_ in set(classes):
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            sa_classes = {"name": class_, "color": hexcolor, "attribute_groups": []}
            classes_loader.append(sa_classes)
        return classes_loader



from typing import Optional

from pydantic import BaseModel
from pydantic import constr
from pydantic import Extra
from superannotate_schemas.schemas.classes import AttributeGroup as AttributeGroupSchema

NotEmptyStr = constr(strict=True, min_length=1)

AttributeGroup = AttributeGroupSchema


class Project(BaseModel):
    name: NotEmptyStr

    class Config:
        extra = Extra.allow


class MLModel(BaseModel):
    name: NotEmptyStr
    id: Optional[int]
    path: NotEmptyStr
    config_path: NotEmptyStr
    team_id: Optional[int]

    class Config:
        extra = Extra.allow


class PriorityScore(BaseModel):
    name: NotEmptyStr
    priority: float



from pathlib import Path

from superannotate.lib.core.enums import AnnotationStatus
from superannotate.lib.core.enums import ImageQuality
from superannotate.lib.core.enums import ProjectState
from superannotate.lib.core.enums import ProjectStatus
from superannotate.lib.core.enums import ProjectType
from superannotate.lib.core.enums import SegmentationStatus
from superannotate.lib.core.enums import TrainingStatus
from superannotate.lib.core.enums import TrainingTask
from superannotate.lib.core.enums import UploadState
from superannotate.lib.core.enums import UserRole


CONFIG_FILE_LOCATION = str(Path.home() / ".superannotate" / "config.json")
LOG_FILE_LOCATION = str(Path.home() / ".superannotate" / "sa.log")
BACKEND_URL = "https://api.annotate.online"

DEFAULT_IMAGE_EXTENSIONS = ["jpg", "jpeg", "png", "tif", "tiff", "webp", "bmp"]
DEFAULT_FILE_EXCLUDE_PATTERNS = ["___save.png", "___fuse.png"]
DEFAULT_VIDEO_EXTENSIONS = ["mp4", "avi", "mov", "webm", "flv", "mpg", "ogg"]
DEFAULT_HYPER_PARAMETERS = {
    "instance_type": "1 x T4 16 GB",
    "num_epochs": 12,
    "dataset_split_ratio": 80,
    "base_lr": 0.02,
    "gamma": 0.5,
    "images_per_batch": 4,
    "batch_per_image": 512,
    "steps": [5],
    "evaluation_period": 12,
    "runtime_seconds": 600,
    "estimated_remaining_time": 600,
    "template_id": None,
}

MODEL_TRAINING_TASKS = {
    "Instance Segmentation for Pixel Projects": "instance_segmentation_pixel",
    "Instance Segmentation for Vector Projects": "instance_segmentation_vector",
    "Keypoint Detection for Vector Projects": "keypoint_detection_vector",
    "Object Detection for Vector Projects": "object_detection_vector",
    "Semantic Segmentation for Pixel Projects": "semantic_segmentation_pixel",
}

AVAILABLE_SEGMENTATION_MODELS = ["autonomous", "generic"]


VECTOR_ANNOTATION_POSTFIX = "___objects.json"
PIXEL_ANNOTATION_POSTFIX = "___pixel.json"
ANNOTATION_MASK_POSTFIX = "___save.png"
ATTACHED_VIDEO_ANNOTATION_POSTFIX = ".json"

NON_PLOTABLE_KEYS = ["eta_seconds", "iteration", "data_time", "time", "model"]

SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES = set('/\\:*?"<>|“')
MAX_PIXEL_RESOLUTION = 4_000_000
MAX_VECTOR_RESOLUTION = 100_000_000
MAX_IMAGE_SIZE = 100 * 1024 * 1024  # 100 MB limit
TOKEN_UUID = "token"

ALREADY_EXISTING_FILES_WARNING = (
    "{} already existing file(s) found that won't be uploaded."
)

ATTACHING_FILES_MESSAGE = "Attaching {} file(s) to project {}."

ATTACHING_UPLOAD_STATE_ERROR = "You cannot attach URLs in this type of project. Please attach it in an external storage project."
UPLOADING_UPLOAD_STATE_ERROR = "You cannot upload files in this type of project. Please upload it in an basic storage project."

INVALID_PROJECT_TYPE_TO_PROCESS = "The function does not support projects of type {}."
DEPRECATED_VIDEO_PROJECTS_MESSAGE = (
    "The function does not support projects containing videos attached with URLs"
)

DEPRECATED_DOCUMENT_PROJECTS_MESSAGE = (
    "The function does not support projects containing documents attached with URLs"
)

LIMITED_FUNCTIONS = {
    ProjectType.VIDEO.value: DEPRECATED_VIDEO_PROJECTS_MESSAGE,
    ProjectType.DOCUMENT.value: DEPRECATED_DOCUMENT_PROJECTS_MESSAGE,
}

DEPRICATED_DOCUMENT_VIDEO_MESSAGE = "The function does not support projects containing videos / documents attached with URLs"

UPLOAD_FOLDER_LIMIT_ERROR_MESSAGE = "The number of items you want to upload exceeds the limit of 50 000 items per folder."
UPLOAD_PROJECT_LIMIT_ERROR_MESSAGE = "The number of items you want to upload exceeds the limit of 500 000 items per project."
UPLOAD_USER_LIMIT_ERROR_MESSAGE = "The number of items you want to upload exceeds the limit of your subscription plan."

ATTACH_FOLDER_LIMIT_ERROR_MESSAGE = "The number of items you want to attach exceeds the limit of 50 000 items per folder."
ATTACH_PROJECT_LIMIT_ERROR_MESSAGE = "The number of items you want to attach exceeds the limit of 500 000 items per project."
ATTACH_USER_LIMIT_ERROR_MESSAGE = "The number of items you want to attach  exceeds the limit of your subscription plan."


COPY_FOLDER_LIMIT_ERROR_MESSAGE = (
    "The number of items you want to copy exceeds the limit of 50 000 items per folder."
)
COPY_PROJECT_LIMIT_ERROR_MESSAGE = "The number of items you want to copy exceeds the limit of 500 000 items per project."
COPY_SUPER_LIMIT_ERROR_MESSAGE = (
    "The number of items you want to copy exceeds the limit of your subscription plan."
)


MOVE_FOLDER_LIMIT_ERROR_MESSAGE = (
    "The number of items you want to move exceeds the limit of 50 000 items per folder."
)
MOVE_PROJECT_LIMIT_ERROR_MESSAGE = "The number of items you want to move exceeds the limit of 500 000 items per project."

PACKAGE_VERSION_INFO_MESSAGE = (
    "Development version {} of SuperAnnotate SDK is being used."
)

PACKAGE_VERSION_MAJOR_UPGRADE = (
    "There is a major upgrade of SuperAnnotate Python SDK available on PyPI. "
    "We recommend upgrading. Run 'pip install --upgrade superannotate' to "
    "upgrade from your version {} to {}."
)

PACKAGE_VERSION_UPGRADE = (
    "There is a newer version of SuperAnnotate Python SDK available on PyPI."
    " Run 'pip install --upgrade superannotate' to"
    " upgrade from your version {} to {}"
)

USE_VALIDATE_MESSAGE = (
    "Use the validate_annotations function to discover the possible reason(s) for "
    "which an annotation is invalid."
)

INVALID_JSON_MESSAGE = "Invalid json"

__alL__ = (
    ProjectState,
    ProjectStatus,
    ProjectType,
    UserRole,
    UploadState,
    TrainingStatus,
    SegmentationStatus,
    TrainingTask,
    ImageQuality,
    AnnotationStatus,
    CONFIG_FILE_LOCATION,
    BACKEND_URL,
    DEFAULT_IMAGE_EXTENSIONS,
    DEFAULT_FILE_EXCLUDE_PATTERNS,
    DEFAULT_VIDEO_EXTENSIONS,
    MAX_IMAGE_SIZE,
    MAX_VECTOR_RESOLUTION,
    MAX_PIXEL_RESOLUTION,
)



import uuid
from functools import wraps
from pathlib import Path
from typing import Optional
from typing import Union

from lib.core.enums import AnnotationStatus
from lib.core.enums import ClassTypeEnum
from lib.core.enums import ProjectStatus
from lib.core.enums import ProjectType
from lib.core.enums import UserRole
from lib.core.exceptions import AppException
from lib.infrastructure.validators import wrap_error
from pydantic import BaseModel
from pydantic import conlist
from pydantic import constr
from pydantic import Extra
from pydantic import Field
from pydantic import parse_obj_as
from pydantic import root_validator
from pydantic import StrictStr
from pydantic import validate_arguments as pydantic_validate_arguments
from pydantic import ValidationError
from pydantic.errors import StrRegexError

NotEmptyStr = constr(strict=True, min_length=1)


class EmailStr(StrictStr):
    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        try:
            constr(
                regex=r"^(?=.{1,254}$)(?=.{1,64}@)[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+)"
                      r"*@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}"
                      r"[a-zA-Z0-9])?)*$"
            ).validate(value)
        except StrRegexError:
            raise ValueError("Invalid email")
        return value


class Status(StrictStr):
    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        if cls.curtail_length and len(value) > cls.curtail_length:
            value = value[: cls.curtail_length]
        if value.lower() not in AnnotationStatus.values():
            raise TypeError(
                f"Available statuses is {', '.join(AnnotationStatus.titles())}. "
            )
        return value


class ProjectStatusEnum(StrictStr):
    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        if cls.curtail_length and len(value) > cls.curtail_length:
            value = value[: cls.curtail_length]
        if value.lower() not in ProjectStatus.values():
            raise TypeError(
                f"Available statuses is {', '.join(ProjectStatus.titles())}. "
            )
        return value


class AnnotatorRole(StrictStr):
    ANNOTATOR_ROLES = (UserRole.ADMIN.name, UserRole.ANNOTATOR.name, UserRole.QA.name)

    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        if cls.curtail_length and len(value) > cls.curtail_length:
            value = value[: cls.curtail_length]
        if value.lower() not in [role.lower() for role in cls.ANNOTATOR_ROLES]:
            raise TypeError(
                f"Invalid user role provided. Please specify one of {', '.join(cls.ANNOTATOR_ROLES)}. "
            )
        return value


class AnnotationType(StrictStr):
    VALID_TYPES = ["bbox", "polygon", "point"]

    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        if value.lower() not in cls.VALID_TYPES:
            raise TypeError(
                f"Available annotation_types are {', '.join(cls.VALID_TYPES)}. "
            )
        return value


class AttachmentDict(BaseModel):
    url: StrictStr
    name: Optional[StrictStr] = Field(default_factory=lambda: str(uuid.uuid4()))

    class Config:
        extra = Extra.ignore

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.url == other.url and self.name.strip() == other.name.strip()


AttachmentArgType = Union[NotEmptyStr, Path, conlist(AttachmentDict, min_items=1)]


class Setting(BaseModel):
    attribute: NotEmptyStr
    value: Union[NotEmptyStr, float, int]

    class Config:
        extra = Extra.ignore


class AttachmentArg(BaseModel):
    __root__: AttachmentArgType

    def __getitem__(self, index):
        return self.__root__[index]

    @property
    def data(self):
        return self.__root__

    @root_validator(pre=True)
    def validate_root(cls, values):
        try:
            parse_obj_as(AttachmentArgType, values["__root__"])
        except ValidationError:
            raise ValueError(
                "The value must be str, path, or list of dicts with the required 'url' and optional 'name' keys"
            )
        return values


class ImageQualityChoices(StrictStr):
    VALID_CHOICES = ["compressed", "original"]

    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        super().validate(value)
        if value.lower() not in cls.VALID_CHOICES:
            raise TypeError(
                f"Image quality available choices are {', '.join(cls.VALID_CHOICES)}."
            )
        return value.lower()


class ProjectTypes(StrictStr):
    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        if value.lower() not in ProjectType.values():
            raise TypeError(
                f" Available project types are {', '.join(ProjectType.titles())}. "
            )
        return value


class ClassType(StrictStr):
    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        enum_values = [e.name.lower() for e in ClassTypeEnum]
        if value.lower() not in enum_values:
            raise TypeError(
                f"Invalid type provided. Please specify one of the {', '.join(enum_values)}. "
            )
        return value.lower()


class AnnotationStatuses(StrictStr):
    @classmethod
    def validate(cls, value: Union[str]) -> Union[str]:
        if value.lower() not in AnnotationStatus.values():
            raise TypeError(
                f"Available an``notation_statuses are {', '.join(AnnotationStatus.titles())}. "
            )
        return value


def validate_arguments(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return pydantic_validate_arguments(func)(*args, **kwargs)
        except ValidationError as e:
            raise AppException(wrap_error(e))

    return wrapped



"""
"""


def _create_vector_instance(
    instance_type,
    points,
    pointLabels,
    attributes,
    class_name="",
    connections=[],
    template_name="",
):
    sa_instance = {
        "type": instance_type,
        "pointLabels": pointLabels,
        "attributes": attributes,
        "creationType": "Preannotation",
    }

    if instance_type == "template":
        sa_instance["points"] = points
        sa_instance["connections"] = connections
        sa_instance["className"] = class_name
        sa_instance["templateName"] = template_name
    elif instance_type == "point":
        sa_instance["x"] = points[0]
        sa_instance["y"] = points[1]
    elif instance_type == "ellipse":
        sa_instance["cx"] = points[0]
        sa_instance["cy"] = points[1]
        sa_instance["rx"] = points[2]
        sa_instance["ry"] = points[3]
        sa_instance["angle"] = points[4]
    elif instance_type == "bbox":
        sa_instance["points"] = {
            "x1": points[0],
            "y1": points[1],
            "x2": points[2],
            "y2": points[3],
        }
    elif instance_type in ("polygon", "polyline"):
        sa_instance["points"] = points
    elif instance_type == "cuboid":
        sa_instance["points"] = {
            "f1": {"x": points[0], "y": points[1]},
            "f2": {"x": points[2], "y": points[3]},
            "r1": {"x": points[4], "y": points[5]},
            "r2": {"x": points[6], "y": points[7]},
        }
    if class_name:
        sa_instance["className"] = class_name

    return sa_instance


def _create_pixel_instance(parts, attributes, class_name=""):
    sa_instance = {
        "attributes": attributes,
        "parts": parts,
    }
    if class_name:
        sa_instance["className"] = class_name

    return sa_instance


def _create_comment(points, comments):
    sa_comment = {
        "type": "comment",
        "x": points[0],
        "y": points[1],
        "correspondence": comments,
    }
    return sa_comment


def _create_sa_json(instances, metadata, tags=[], comments=[]):
    return {
        "instances": instances,
        "metadata": metadata,
        "tags": tags,
        "comments": comments,
    }



"""
VOC to SA conversion method
"""
import threading

import cv2
import numpy as np
from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance
from .voc_helper import _get_image_shape_from_xml
from .voc_helper import _get_voc_instances_from_xml
from .voc_helper import _iou

logger = get_default_logger()


def _generate_polygons(object_mask_path):
    segmentation = []

    object_mask = cv2.imread(str(object_mask_path), cv2.IMREAD_GRAYSCALE)

    object_unique_colors = np.unique(object_mask)

    index = 1
    groupId = 0
    for unique_color in object_unique_colors:
        if unique_color in (0, 220):
            continue

        mask = np.zeros_like(object_mask)
        mask[object_mask == unique_color] = 255
        contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        segment = []
        if len(contours) > 1:
            for contour in contours:
                segment.append(contour.flatten().tolist())
            groupId = index
            index += 1
        else:
            segment.append(contours[0].flatten().tolist())
            groupId = 0

        segmentation.append((segment, groupId))

    return segmentation


def _generate_instances(polygon_instances, voc_instances):
    instances = []
    for polygon, group_id in polygon_instances:
        ious = []
        if len(polygon) > 1:
            temp = []
            for poly in polygon:
                temp += poly
        else:
            temp = polygon[0]
        bbox_poly = [min(temp[::2]), min(temp[1::2]), max(temp[::2]), max(temp[1::2])]
        for _, bbox in voc_instances:
            ious.append(_iou(bbox_poly, bbox))
        ind = np.argmax(ious)
        for poly in polygon:
            class_name = list(voc_instances[ind][0].keys())[0]
            attributes = voc_instances[ind][0][class_name]
            instances.append(
                {
                    "className": class_name,
                    "polygon": poly,
                    "bbox": voc_instances[ind][1],
                    "groupId": group_id,
                    "classAttributes": attributes,
                }
            )
    return instances


def voc_instance_segmentation_to_sa_vector(voc_root, output_dir):
    classes = []
    object_masks_dir = voc_root / "SegmentationObject"
    annotation_dir = voc_root / "Annotations"
    file_list = list(object_masks_dir.glob("*"))
    if not file_list:
        logger.warning(
            "You need to have both 'Annotations' and 'SegmentationObject' directories to be able to convert."
        )

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(file_list), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for filename in file_list:
        polygon_instances = _generate_polygons(object_masks_dir / filename.name)
        voc_instances = _get_voc_instances_from_xml(annotation_dir / filename.name)
        for class_, _ in voc_instances:
            classes.append(class_)

        maped_instances = _generate_instances(polygon_instances, voc_instances)
        sa_instances = []
        for instance in maped_instances:
            sa_obj = _create_vector_instance(
                "polygon",
                instance["polygon"],
                {},
                instance["classAttributes"],
                instance["className"],
            )
            sa_instances.append(sa_obj)

        images_converted.append(filename)
        file_name = "%s.jpg___objects.json" % filename.stem
        height, width = _get_image_shape_from_xml(annotation_dir / filename.name)
        sa_metadata = {"name": filename.stem, "height": height, "width": width}
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)

    finish_event.set()
    tqdm_thread.join()
    return classes


def voc_object_detection_to_sa_vector(voc_root, output_dir):
    classes = []
    annotation_dir = voc_root / "Annotations"
    file_list = list(annotation_dir.glob("*"))
    if not file_list:
        logger.warning("'Annotations' directory is empty")

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(file_list), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()

    for filename in file_list:
        voc_instances = _get_voc_instances_from_xml(annotation_dir / filename.name)
        sa_instances = []
        for class_, bbox in voc_instances:
            class_name = list(class_.keys())[0]
            classes.append(class_)

            points = (bbox[0], bbox[1], bbox[2], bbox[3])
            sa_obj = _create_vector_instance(
                "bbox", points, {}, class_[class_name], class_name
            )
            sa_instances.append(sa_obj)

        images_converted.append(filename)
        file_name = "%s.jpg___objects.json" % filename.stem
        height, width = _get_image_shape_from_xml(annotation_dir / filename.name)
        sa_metadata = {"name": filename.stem, "height": height, "width": width}
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)

    finish_event.set()
    tqdm_thread.join()
    return classes



import random
import time
from abc import ABCMeta
from abc import abstractmethod
from collections import defaultdict
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import List

import lib.core as constances
from lib.core.enums import ClassTypeEnum
from lib.core.reporter import Reporter
from superannotate_schemas.schemas.classes import AnnotationClass
from superannotate_schemas.schemas.classes import Attribute
from superannotate_schemas.schemas.classes import AttributeGroup


class BaseDataHandler(metaclass=ABCMeta):
    @abstractmethod
    def handle(self, *args, **kwargs):
        raise NotImplementedError


class ClassIdGenerator:
    def __init__(self):
        self.classes = defaultdict(int)
        self.idx = -1

    def send(self, class_name: str):
        if class_name not in self.classes:
            self.classes[class_name] = self.idx
            self.idx -= 1
        return self.classes[class_name]


class BaseAnnotationDateHandler(BaseDataHandler, metaclass=ABCMeta):
    def __init__(self, annotation_classes: List[AnnotationClass]):
        self._annotation_classes: List[AnnotationClass] = annotation_classes

    @lru_cache()
    def get_annotation_class(self, name: str) -> AnnotationClass:
        for annotation_class in self._annotation_classes:
            if annotation_class.name == name:
                return annotation_class

    @lru_cache()
    def get_attribute_group(
        self, annotation_class: AnnotationClass, attr_group_name: str
    ) -> AttributeGroup:
        for attr_group in annotation_class.attribute_groups:
            if attr_group.name == attr_group_name:
                return attr_group

    @lru_cache()
    def get_attribute(self, attr_group: AttributeGroup, attr_name: str) -> Attribute:
        for attr in attr_group.attributes:
            if attr.name == attr_name:
                return attr


class ChainedAnnotationHandlers:
    def __init__(self):
        self._handlers: List[BaseDataHandler] = list()

    def handle(self, data: Any):
        for handler in self._handlers:
            data = handler.handle(data)
        return data

    def attach(self, handler: BaseDataHandler):
        if handler not in self._handlers:
            self._handlers.append(handler)

    def detach(self, handler):
        if handler in self._handlers:
            self._handlers.remove(handler)

    def detach_all(self):
        self._handlers.clear()


class AnnotationStatusHandler(BaseDataHandler):
    def handle(self, annotations: Dict):
        annotations["metadata"][
            "status"
        ] = constances.AnnotationStatus.IN_PROGRESS.value
        return annotations


class LastActionHandler(BaseDataHandler):
    def __init__(self, email: str):
        self._email = email

    def handle(self, annotations: Dict):
        annotations["metadata"]["lastAction"] = {
            "email": self._email,
            "timestamp": int(time.time()),
        }
        return annotations


class DocumentTagHandler(BaseAnnotationDateHandler):
    def handle(self, annotation: dict):
        new_tags = []
        for tag in annotation["tags"]:
            annotation_class = self.get_annotation_class(tag)
            if annotation_class:
                new_tags.append(annotation_class.id)
        annotation["tags"] = new_tags
        return annotation


class MissingIDsHandler(BaseAnnotationDateHandler):
    def __init__(
        self,
        annotation_classes: List[AnnotationClass],
        templates: List[dict],
        reporter: Reporter,
    ):
        super().__init__(annotation_classes)
        self.validate_existing_classes(annotation_classes)
        self._templates = templates
        self.reporter = reporter

    def validate_existing_classes(self, annotation_classes: List[AnnotationClass]):
        classes_data = defaultdict(dict)
        for annotation_class in annotation_classes:
            class_info = {"id": annotation_class.id, "attribute_groups": {}}
            if annotation_class.attribute_groups:
                for attribute_group in annotation_class.attribute_groups:
                    attribute_group_data = defaultdict(dict)
                    for attribute in attribute_group.attributes:
                        if attribute.name in attribute_group_data.keys():
                            self.reporter.log_warning(
                                f"Duplicate annotation class attribute name {attribute.name}"
                                f" in attribute group {attribute_group.name}. "
                                "Only one of the annotation class attributes will be used. "
                                "This will result in errors in annotation upload."
                            )
                        attribute_group_data[attribute.name] = attribute.id
                    if attribute_group.name in class_info["attribute_groups"].keys():
                        self.reporter.log_warning(
                            f"Duplicate annotation class attribute group name {attribute_group.name}."
                            " Only one of the annotation class attribute groups will be used."
                            " This will result in errors in annotation upload."
                        )
                    class_info["attribute_groups"][attribute_group.name] = {
                        "id": attribute_group.id,
                        "attributes": attribute_group_data,
                    }
            if annotation_class.name in classes_data.keys():
                self.reporter.log_warning(
                    f"Duplicate annotation class name {annotation_class.name}."
                    f" Only one of the annotation classes will be used."
                    " This will result in errors in annotation upload.",
                )

    def _get_class_type(self, annotation_type: str):
        if annotation_type == ClassTypeEnum.TAG.name:
            return ClassTypeEnum.TAG
        return ClassTypeEnum.OBJECT

    def handle(self, annotation: dict):
        if "instances" not in annotation:
            return annotation
        for annotation_instance in annotation["instances"]:
            if "className" not in annotation_instance:
                annotation_instance["classId"] = -1
            else:
                class_name = annotation_instance["className"]
                annotation_class = self.get_annotation_class(class_name)
                if not annotation_class:
                    self.reporter.log_warning(f"Couldn't find class {class_name}")
                    self.reporter.store_message("missing_classes", class_name)
                    annotation_instance["classId"] = -1
                    annotation_instance["attributes"] = []
                    self._annotation_classes.append(
                        AnnotationClass(
                            id=-1,
                            name=class_name,
                            attribute_groups=[],
                            color=f"#{random.randint(0, 0xFFFFFF):06x}",
                        )
                    )
                    self.get_annotation_class.cache_clear()
                else:
                    annotation_instance["classId"] = annotation_class.id

        template_name_id_map = {
            template["name"]: template["id"] for template in self._templates
        }
        for annotation_instance in (
            i for i in annotation["instances"] if i.get("type", None) == "template"
        ):
            annotation_instance["templateId"] = template_name_id_map.get(
                annotation_instance.get("templateName", ""), -1
            )

        for annotation_instance in [
            i for i in annotation["instances"] if "className" in i and i["classId"] > 0
        ]:
            annotation_class_name = annotation_instance["className"]
            annotation_class = self.get_annotation_class(annotation_class_name)

            if not annotation_class:
                self.reporter.log_warning(
                    f"Couldn't find annotation class {annotation_class_name}"
                )
                continue
            annotation_instance_attributes = []
            for annotation_attribute in annotation_instance["attributes"]:
                attr_group = self.get_attribute_group(
                    annotation_class, annotation_attribute["groupName"]
                )
                if not attr_group:
                    self.reporter.log_warning(
                        f"Couldn't find annotation group {annotation_attribute['groupName']}."
                    )
                    self.reporter.store_message(
                        "missing_attribute_groups",
                        f"{annotation_instance['className']}.{annotation_attribute['groupName']}",
                    )
                    continue
                annotation_attribute["groupId"] = attr_group.id
                attribute = self.get_attribute(attr_group, annotation_attribute["name"])
                if not attribute:
                    del annotation_attribute["groupId"]
                    self.reporter.log_warning(
                        f"Couldn't find annotation name {annotation_attribute['name']} in"
                        f" annotation group {annotation_attribute['groupName']}",
                    )
                    self.reporter.store_message(
                        "missing_attributes", annotation_attribute["name"]
                    )
                    continue
                annotation_attribute["id"] = attribute.id
                annotation_instance_attributes.append(annotation_attribute)
            annotation_instance["attributes"] = annotation_instance_attributes
        return annotation


class VideoFormatHandler(BaseAnnotationDateHandler):
    def __init__(self, annotation_classes: List[AnnotationClass], reporter: Reporter):
        super().__init__(annotation_classes)
        self.reporter = reporter

    def handle(self, annotation: dict):
        id_generator = ClassIdGenerator()

        def safe_time(timestamp):
            return "0" if str(timestamp) == "0.0" else timestamp

        def convert_timestamp(timestamp):
            return timestamp / 10 ** 6 if timestamp else "0"

        editor_data = {
            "instances": [],
            "tags": annotation["tags"],
            "name": annotation["metadata"]["name"],
            "metadata": {
                "name": annotation["metadata"]["name"],
                "width": annotation["metadata"].get("width"),
                "height": annotation["metadata"].get("height"),
            },
        }
        if annotation["metadata"].get("duration"):
            editor_data["metadata"]["duration"] = convert_timestamp(
                annotation["metadata"]["duration"]
            )
        for instance in annotation["instances"]:
            meta = instance["meta"]
            class_name = meta.get("className")
            editor_instance = {
                "attributes": [],
                "timeline": {},
                "type": meta["type"],
                "locked": False,
            }
            if class_name:
                annotation_class = self.get_annotation_class(class_name)
                if annotation_class:
                    editor_instance["classId"] = annotation_class.id
                else:
                    editor_instance["classId"] = id_generator.send(class_name)
            else:
                editor_instance["classId"] = id_generator.send("unknown_class")

            if meta.get("pointLabels", None):
                editor_instance["pointLabels"] = meta["pointLabels"]

            active_attributes = set()
            for parameter in instance["parameters"]:

                start_time = safe_time(convert_timestamp(parameter["start"]))
                end_time = safe_time(convert_timestamp(parameter["end"]))

                for timestamp_data in parameter["timestamps"]:
                    timestamp = safe_time(
                        convert_timestamp(timestamp_data["timestamp"])
                    )
                    editor_instance["timeline"][timestamp] = {}

                    if timestamp == start_time:
                        editor_instance["timeline"][timestamp]["active"] = True

                    if timestamp == end_time:
                        editor_instance["timeline"][timestamp]["active"] = False

                    if timestamp_data.get("points", None):
                        editor_instance["timeline"][timestamp][
                            "points"
                        ] = timestamp_data["points"]
                    if not class_name:
                        continue
                    annotation_class = self.get_annotation_class(class_name)
                    if not annotation_class:
                        self.reporter.store_message(
                            "missing_classes", meta["className"]
                        )
                        continue

                    existing_attributes_in_current_instance = set()
                    for attribute in timestamp_data["attributes"]:
                        group_name, attr_name = (
                            attribute.get("groupName"),
                            attribute.get("name"),
                        )
                        attr_group = self.get_attribute_group(
                            annotation_class, group_name
                        )
                        attribute = (
                            self.get_attribute(attr_group, attr_name)
                            if attr_group
                            else None
                        )
                        if not attr_group:
                            self.reporter.store_message(
                                "missing_attribute_groups", f"{class_name}.{group_name}"
                            )
                        elif not attribute:
                            self.reporter.store_message(
                                "missing_attributes",
                                f"{class_name}.{group_name}.{attr_name}",
                            )
                        else:
                            existing_attributes_in_current_instance.add(
                                (group_name, attr_name)
                            )
                    attributes_to_add = (
                        existing_attributes_in_current_instance - active_attributes
                    )
                    attributes_to_delete = (
                        active_attributes - existing_attributes_in_current_instance
                    )
                    if attributes_to_add or attributes_to_delete:
                        editor_instance["timeline"][timestamp][
                            "attributes"
                        ] = defaultdict(list)
                    for new_attribute in attributes_to_add:
                        attribute_group = self.get_attribute_group(
                            annotation_class, new_attribute[0]
                        )
                        attribute = self.get_attribute(
                            attribute_group, new_attribute[1]
                        )
                        attr = {
                            "id": attribute.id,
                            "groupId": attribute_group.id,
                        }
                        active_attributes.add(new_attribute)
                        editor_instance["timeline"][timestamp]["attributes"][
                            "+"
                        ].append(attr)
                    for attribute_to_delete in attributes_to_delete:
                        attribute_group = self.get_attribute_group(
                            annotation_class, attribute_to_delete[0]
                        )
                        attribute = self.get_attribute(
                            attribute_group, attribute_to_delete[1]
                        )
                        attr = {
                            "id": attribute.id,
                            "groupId": attribute_group.id,
                        }
                        active_attributes.remove(attribute_to_delete)
                        editor_instance["timeline"][timestamp]["attributes"][
                            "-"
                        ].append(attr)
            editor_data["instances"].append(editor_instance)
        return editor_data



"""
YOLO to SA conversion method
"""
import threading
from glob import glob
from pathlib import Path

import cv2
from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance

logger = get_default_logger()


def yolo_object_detection_to_sa_vector(data_path, output_dir):
    classes = {}
    id_ = 0
    classes_file = open(data_path / "classes.txt")
    for line in classes_file:
        key = line.rstrip()
        if key not in classes.keys():
            classes[id_] = key
            id_ += 1

    annotations = [
        annot for annot in data_path.glob("*.txt") if annot.name != "classes.txt"
    ]

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(annotations), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for annotation in annotations:
        file = open(annotation)
        file_name = "%s.*" % annotation.stem
        files_list = glob(str(data_path / file_name))
        if len(files_list) == 1:
            images_not_converted.append(annotation.name)
            logger.warning("'%s' image for annotation doesn't exist", annotation)
            continue
        if len(files_list) > 2:
            images_not_converted.append(annotation.name)
            logger.warning("'%s' multiple file for this annotation", annotation)
            continue

        if Path(files_list[0]).suffix == ".txt":
            file_name = files_list[1]
        else:
            file_name = files_list[0]

        img = cv2.imread(file_name)
        H, W, _ = img.shape

        sa_instances = []
        for line in file:
            values = line.split()
            class_id = int(values[0])
            points = (
                float(values[1]) * W - float(values[3]) * W / 2,
                float(values[2]) * H - float(values[4]) * H / 2,
                float(values[1]) * W + float(values[3]) * W / 2,
                float(values[2]) * H + float(values[4]) * H / 2,
            )
            sa_obj = _create_vector_instance("bbox", points, {}, [], classes[class_id])
            sa_instances.append(sa_obj.copy())

        images_converted.append(annotation.name)
        file_name = "%s___objects.json" % Path(file_name).name
        sa_metadata = {"name": Path(file_name).name, "width": W, "height": H}
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)

    finish_event.set()
    tqdm_thread.join()
    return classes



from pathlib import Path

import pandas as pd
import plotly.express as px
from lib.app.mixp.decorators import Trackable
from superannotate.lib.app.exceptions import AppException
from superannotate.lib.core import DEPRICATED_DOCUMENT_VIDEO_MESSAGE
from superannotate.logger import get_default_logger

from .common import aggregate_image_annotations_as_df

logger = get_default_logger()


@Trackable
def class_distribution(export_root, project_names, visualize=False):
    """Aggregate distribution of classes across multiple projects.

    :param export_root: root export path of the projects
    :type export_root: Pathlike (str or Path)
    :param project_names: list of project names to aggregate through
    :type project_names: list of str
    :param visualize: enables class histogram plot
    :type visualize: bool
    :return: DataFrame on class distribution with columns ["className", "count"]
    :rtype: pandas DataFrame
    """

    json_paths = list(Path(str(export_root)).glob("*.json"))
    if (
        json_paths
        and "___pixel.json" not in json_paths[0].name
        and "___objects.json" not in json_paths[0].name
    ):
        raise AppException(DEPRICATED_DOCUMENT_VIDEO_MESSAGE)

    logger.info(
        "Aggregating class distribution accross projects: {}.".format(
            " ".join(project_names)
        ),
    )

    project_df_list = []
    for project_name in project_names:
        project_root = Path(export_root).joinpath(project_name)
        project_df = aggregate_image_annotations_as_df(
            project_root, include_classes_wo_annotations=True
        )
        project_df = project_df[["imageName", "instanceId", "className"]]
        project_df["projectName"] = project_name
        project_df_list.append(project_df)

    df = pd.concat(project_df_list, ignore_index=True)

    df["id"] = (
        df["projectName"] + "_" + df["imageName"] + "_" + df["instanceId"].astype(str)
    )
    df = df.groupby("className")["id"].nunique()
    df = df.reset_index().rename(columns={"id": "count"})
    df = df.sort_values(["count"], ascending=False)

    if visualize:
        fig = px.bar(df, x="className", y="count",)
        fig.update_traces(hovertemplate="%{x}: %{y}")
        fig.update_yaxes(title_text="Instance Count")
        fig.update_xaxes(title_text="")
        fig.show()

    return df



from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple

from lib.core.reporter import Reporter
from lib.core.service_types import ServiceResponse


class SuperannotateServiceProvider:
    @abstractmethod
    def attach_files(
        self,
        project_id: int,
        folder_id: int,
        team_id: int,
        files: List[Dict],
        annotation_status_code: int,
        upload_state_code: int,
        meta: Dict,
    ):
        raise NotImplementedError

    @abstractmethod
    def get_annotation_classes(
        self, project_id: int, team_id: int, name_prefix: str = None
    ):
        raise NotImplementedError

    @abstractmethod
    def share_project_bulk(self, project_id: int, team_id: int, users: Iterable):
        raise NotImplementedError

    @abstractmethod
    def invite_contributors(
        self, team_id: int, team_role: int, emails: Iterable
    ) -> Tuple[List[str], List[str]]:
        raise NotImplementedError

    @abstractmethod
    def prepare_export(
        self,
        project_id: int,
        team_id: int,
        folders: List[str],
        annotation_statuses: Iterable[Any],
        include_fuse: bool,
        only_pinned: bool,
    ):
        raise NotImplementedError

    @abstractmethod
    def delete_team_invitation(self, team_id: int, token: str, email: str):
        raise NotImplementedError

    @abstractmethod
    def search_team_contributors(self, team_id: int, query_string: str = None):
        raise NotImplementedError

    @abstractmethod
    def get_project_settings(self, project_id: int, team_id: int):
        raise NotImplementedError

    @abstractmethod
    def set_project_settings(self, project_id: int, team_id: int, data: List):
        raise NotImplementedError

    @abstractmethod
    def get_project_workflows(self, project_id: int, team_id: int):
        raise NotImplementedError

    @abstractmethod
    def list_images(self, query_string):
        raise NotImplementedError

    @abstractmethod
    def get_project(self, uuid: int, team_id: int):
        raise NotImplementedError

    @abstractmethod
    def set_project_workflow(self, project_id: int, team_id: int, data: Dict):
        raise NotImplementedError

    def delete_folders(self, project_id: int, team_id: int, folder_ids: List[int]):
        raise NotImplementedError

    def get_folder(self, query_string: str):
        raise NotImplementedError

    def get_folders(self, query_string: str = None, params: dict = None):
        raise NotImplementedError

    def create_folder(self, project_id: int, team_id: int, folder_name: str):
        raise NotImplementedError

    def update_folder(self, project_id: int, team_id: int, folder_data: dict):
        raise NotImplementedError

    def get_download_token(
        self,
        project_id: int,
        team_id: int,
        folder_id: int,
        image_id: int,
        include_original: int = 1,
    ) -> dict:
        raise NotImplementedError

    def get_upload_token(
        self, project_id: int, team_id: int, folder_id: int, image_id: int,
    ) -> dict:
        raise NotImplementedError

    def update_image(self, image_id: int, team_id: int, project_id: int, data: dict):
        raise NotImplementedError

    def copy_items_between_folders_transaction(
        self,
        team_id: int,
        project_id: int,
        from_folder_id: int,
        to_folder_id: int,
        items: List[str],
        include_annotations: bool = False,
        include_pin: bool = False,
    ) -> int:
        raise NotImplementedError

    def move_images_between_folders(
        self,
        team_id: int,
        project_id: int,
        from_folder_id: int,
        to_folder_id: int,
        images: List[str],
    ) -> List[str]:
        """
        Returns list of moved images.
        """
        raise NotImplementedError

    def get_duplicated_images(
        self, project_id: int, team_id: int, folder_id: int, images: List[str]
    ):
        raise NotImplementedError

    def get_progress(
        self, project_id: int, team_id: int, poll_id: int
    ) -> Tuple[int, int]:
        raise NotImplementedError

    def await_progress(self, project_id: int, team_id: int, poll_id: int, items_count) -> Tuple[int, int]:
        raise NotImplementedError

    def set_images_statuses_bulk(
        self,
        image_names: List[str],
        team_id: int,
        project_id: int,
        folder_id: int,
        annotation_status: int,
    ):
        raise NotImplementedError

    def delete_images(self, project_id: int, team_id: int, image_ids: List[int]):
        raise NotImplementedError

    def assign_images(
        self,
        team_id: int,
        project_id: int,
        folder_name: str,
        user: str,
        image_names: list,
    ):
        raise NotImplementedError

    def get_bulk_images(
        self, project_id: int, team_id: int, folder_id: int, images: List[str]
    ) -> List[dict]:
        raise NotImplementedError

    def un_assign_folder(
        self, team_id: int, project_id: int, folder_name: str,
    ):
        raise NotImplementedError

    def assign_folder(
        self, team_id: int, project_id: int, folder_name: str, users: list
    ):
        raise NotImplementedError

    def un_assign_images(
        self, team_id: int, project_id: int, folder_name: str, image_names: list,
    ):
        raise NotImplementedError

    def un_share_project(
        self, team_id: int, project_id: int, user_id: str,
    ):
        raise NotImplementedError

    def upload_form_s3(
        self,
        project_id: int,
        team_id: int,
        access_key: str,
        secret_key: str,
        bucket_name: str,
        from_folder_name: str,
        to_folder_id: int,
    ):
        raise NotImplementedError

    def get_upload_status(self, project_id: int, team_id: int, folder_id: int):
        raise NotImplementedError

    def get_exports(self, team_id: int, project_id: int):
        raise NotImplementedError

    def get_export(self, team_id: int, project_id: int, export_id: int):
        raise NotImplementedError

    def get_project_images_count(self, team_id: int, project_id: int):
        raise NotImplementedError

    def get_s3_upload_auth_token(self, team_id: int, folder_id: int, project_id: int):
        raise NotImplementedError

    def delete_annotation_class(
        self, team_id: int, project_id: int, annotation_class_id: int
    ):
        raise NotImplementedError

    def set_annotation_classes(self, team_id: int, project_id: int, data: list):
        raise NotImplementedError

    def set_project_workflow_bulk(self, project_id: int, team_id: int, steps: list):
        raise NotImplementedError

    def set_project_workflow_attributes_bulk(
        self, project_id: int, team_id: int, attributes: list
    ):
        raise NotImplementedError

    def get_pre_annotation_upload_data(
        self, project_id: int, team_id: int, image_ids: List[int], folder_id: int
    ):
        raise NotImplementedError

    def get_annotation_upload_data(
        self, project_id: int, team_id: int, image_ids: List[int], folder_id: int
    ) -> ServiceResponse:
        raise NotImplementedError

    def get_templates(self, team_id: int):
        raise NotImplementedError

    def start_model_training(self, team_id: int, hyper_parameters: dict) -> dict:
        raise NotImplementedError

    def get_model_metrics(self, team_id: int, model_id: int) -> dict:
        raise NotImplementedError

    def get_models(
        self, name: str, team_id: int, project_id: int, model_type: str
    ) -> List:
        raise NotImplementedError

    def bulk_get_folders(self, team_id: int, project_ids: List[int]):
        raise NotImplementedError

    def update_model(self, team_id: int, model_id: int, data: dict):
        raise NotImplementedError

    def delete_model(self, team_id: int, model_id: int):
        raise NotImplementedError

    def get_ml_model_download_tokens(
        self, team_id: int, model_id: int
    ) -> ServiceResponse:
        raise NotImplementedError

    def run_prediction(
        self, team_id: int, project_id: int, ml_model_id: int, image_ids: list
    ):
        raise NotImplementedError

    def delete_image_annotations(
        self,
        team_id: int,
        project_id: int,
        folder_id: int = None,
        image_names: List[str] = None,
    ) -> dict:
        raise NotImplementedError

    def get_annotations_delete_progress(
        self, team_id: int, project_id: int, poll_id: int
    ):
        raise NotImplementedError

    def get_limitations(
        self, team_id: int, project_id: int, folder_id: int = None
    ) -> ServiceResponse:
        raise NotImplementedError

    @abstractmethod
    def get_annotations(
        self,
        project_id: int,
        team_id: int,
        folder_id: int,
        items: List[str],
        reporter: Reporter,
    ) -> List[dict]:
        raise NotImplementedError

    def upload_priority_scores(
        self, team_id: int, project_id: int, folder_id: int, priorities: list
    ) -> dict:
        raise NotImplementedError

    def get_integrations(self, team_id: int) -> List[dict]:
        raise NotImplementedError

    def attach_integrations(
        self,
        team_id: int,
        project_id: int,
        integration_id: int,
        folder_id: int,
        folder_name: str,
    ) -> bool:
        raise NotImplementedError

    def saqul_query(
        self, team_id: int, project_id: int, query: str, folder_id: int
    ) -> ServiceResponse:
        raise NotImplementedError

    def validate_saqul_query(self, team_id: int, project_id: int, query: str) -> dict:
        raise NotImplementedError



from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
from typing import Iterable
from typing import List

from lib.core.exceptions import AppValidationException
from lib.core.reporter import Reporter
from lib.core.response import Response


class BaseUseCase(ABC):
    def __init__(self):
        self._response = Response()
        self._pass_validation = False

    @abstractmethod
    def execute(self) -> Response:
        raise NotImplementedError

    def _validate(self):
        for name in dir(self):
            try:
                if name.startswith("validate_"):
                    method = getattr(self, name)
                    method()
            except AppValidationException as e:
                self._response.errors = e

    def is_valid(self):
        if self._pass_validation:
            return True
        self._validate()
        return not self._response.errors


class BaseInteractiveUseCase(BaseUseCase):
    def __init__(self):
        super().__init__()
        self._validated = False

    def is_valid(self):
        if not self._validated:
            self._validate()
            self._validated = True
        return not self._response.errors

    @property
    def response(self):
        return self._response

    @property
    def data(self):
        return self.response.data

    @abstractmethod
    def execute(self) -> Iterable:
        raise NotImplementedError


class BaseReportableUseCae(BaseUseCase, metaclass=ABCMeta):
    def __init__(self, reporter: Reporter):
        super().__init__()
        self.reporter = reporter


class BaseUserBasedUseCase(BaseReportableUseCae, metaclass=ABCMeta):
    """
    class contain validation of unique emails
    """

    def __init__(self, reporter: Reporter, emails: List[str]):
        super().__init__(reporter)
        self._emails = emails

    def validate_emails(self):
        emails_to_add = set()
        unique_emails = [
            email
            for email in self._emails
            if email not in emails_to_add and not emails_to_add.add(email)
        ]
        if unique_emails:
            self.reporter.log_info(
                f"Dropping duplicates. Found {len(unique_emails)}/{len(self._emails)} unique users."
            )
        self._emails = emails_to_add



import numpy as np

from ....common import write_to_json
from ..baseStrategy import baseStrategy


class VocStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        classes = self.conversion_algorithm(self.export_root, self.output_dir)
        sa_classes = self._create_classes(classes)
        (self.output_dir / "classes").mkdir(exist_ok=True)
        write_to_json(self.output_dir / "classes" / "classes.json", sa_classes)

        # if self.project_type == 'Pixel':
        #     all_files = self.output_dir.glob('*.png')
        #     for file in all_files:
        #         if '___save.png' not in str(file.name):
        #             (self.output_dir / file.name).unlink()
        all_files = self.output_dir.glob("*.png")
        for file in all_files:
            if "___save.png" not in str(file.name):
                (self.output_dir / file.name).unlink()

    def _create_classes(self, instances):
        classes = {}
        for instance in instances:
            for class_, value in instance.items():
                if class_ not in classes:
                    classes[class_] = {}

                for attr in value:
                    if attr["groupName"] in classes[class_]:
                        classes[class_][attr["groupName"]].append(attr["name"])
                    else:
                        classes[class_][attr["groupName"]] = [attr["name"]]

        sa_classes = []
        for class_ in classes:
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            attribute_groups = []
            for attr_group, value in classes[class_].items():
                attributes = []
                for attr in set(value):
                    attributes.append({"name": attr})

                attribute_groups.append(
                    {"name": attr_group, "is_multiselect": 0, "attributes": attributes}
                )
            sa_class = {
                "name": class_,
                "color": hexcolor,
                "attribute_groups": attribute_groups,
            }
            sa_classes.append(sa_class)
        return sa_classes



from collections import namedtuple
from typing import Any
from typing import List
from typing import NamedTuple

CONDITION_OR = "|"
CONDITION_AND = "&"
CONDITION_EQ = "="
CONDITION_GT = ">"
CONDITION_GE = ">="
CONDITION_LT = "<"
CONDITION_LE = "<="


class EmptyCondition:
    def __or__(self, other):
        return other

    def __and__(self, other):
        return other

    def build_query(self):
        return ""


class Condition:
    def __init__(self, key: str, value: Any, condition_type: str):
        self._key = key
        self._value = value
        self._type = condition_type
        self._condition_set = []  # type: List[NamedTuple]

    @staticmethod
    def get_empty_condition():
        return EmptyCondition()

    def __str__(self):
        return f"{self._key}{self._type}{self._value}"

    def __or__(self, other):
        if not isinstance(other, Condition):
            raise Exception("Support the only Condition types")
        QueryCondition = namedtuple("QueryCondition", ("condition", "query"))
        self._condition_set.append(QueryCondition(CONDITION_OR, other.build_query()))
        return self

    def __and__(self, other):
        if isinstance(other, tuple) or isinstance(other, list):
            for elem in other:
                if isinstance(other, EmptyCondition):
                    continue
                if not isinstance(other, Condition):
                    raise Exception("Support the only Condition types")
                return self.__and__(elem)
        elif not isinstance(other, (Condition, EmptyCondition)):
            raise Exception("Support the only Condition types")
        QueryCondition = namedtuple("QueryCondition", ("condition", "query"))
        self._condition_set.append(QueryCondition(CONDITION_AND, other.build_query()))
        return self

    def build_query(self):
        return str(self) + "".join(
            [f"{condition[0]}{condition[1]}" for condition in self._condition_set]
        )



import json
from typing import Callable

import aiohttp
from lib.core.reporter import Reporter


class StreamedAnnotations:
    DELIMITER = b"\\n;)\\n"

    def __init__(self, headers: dict, reporter: Reporter):
        self._headers = headers
        self._annotations = []
        self._reporter = reporter

    async def fetch(
        self,
        method: str,
        session: aiohttp.ClientSession,
        url: str,
        data: dict = None,
        params: dict = None,
    ):
        response = await session._request(method, url, json=data, params=params)
        buffer = b""
        async for line in response.content.iter_any():
            slices = line.split(self.DELIMITER)
            if len(slices) == 1:
                buffer += slices[0]
                continue
            elif slices[0]:
                self._annotations.append(json.loads(buffer + slices[0]))
                self._reporter.update_progress()
            for data in slices[1:-1]:
                self._annotations.append(json.loads(data))
                self._reporter.update_progress()
            buffer = slices[-1]
        if buffer:
            self._annotations.append(json.loads(buffer))
            self._reporter.update_progress()
        return self._annotations

    async def get_data(
        self,
        url: str,
        data: list,
        method: str = "post",
        params=None,
        chunk_size: int = 100,
        map_function: Callable = lambda x: x,
        verify_ssl: bool = False,
    ):
        async with aiohttp.ClientSession(
            raise_for_status=True,
            headers=self._headers,
            connector=aiohttp.TCPConnector(ssl=verify_ssl),
        ) as session:

            if chunk_size:
                for i in range(0, len(data), chunk_size):
                    data_to_process = data[i : i + chunk_size]
                    await self.fetch(
                        method,
                        session,
                        url,
                        map_function(data_to_process),
                        params=params,
                    )
            else:
                await self.fetch(
                    method, session, url, map_function(data), params=params
                )
        return self._annotations



import concurrent.futures
import os.path
import tempfile
import time
import zipfile
from pathlib import Path
from typing import List

import boto3
import lib.core as constances
import pandas as pd
import requests
from botocore.exceptions import ClientError
from lib.app.analytics.common import aggregate_image_annotations_as_df
from lib.app.analytics.common import consensus_plot
from lib.app.analytics.common import image_consensus
from lib.core.conditions import Condition
from lib.core.conditions import CONDITION_EQ as EQ
from lib.core.entities import FolderEntity
from lib.core.entities import MLModelEntity
from lib.core.entities import ProjectEntity
from lib.core.enums import ExportStatus
from lib.core.exceptions import AppException
from lib.core.exceptions import AppValidationException
from lib.core.repositories import BaseManageableRepository
from lib.core.serviceproviders import SuperannotateServiceProvider
from lib.core.usecases.base import BaseInteractiveUseCase
from lib.core.usecases.base import BaseUseCase
from lib.core.usecases.images import GetBulkImages
from superannotate.logger import get_default_logger


logger = get_default_logger()


class PrepareExportUseCase(BaseUseCase):
    def __init__(
        self,
        project: ProjectEntity,
        folder_names: List[str],
        backend_service_provider: SuperannotateServiceProvider,
        include_fuse: bool,
        only_pinned: bool,
        annotation_statuses: List[str] = None,
    ):
        super().__init__(),
        self._project = project
        self._folder_names = list(folder_names) if folder_names else None
        self._backend_service = backend_service_provider
        self._annotation_statuses = annotation_statuses
        self._include_fuse = include_fuse
        self._only_pinned = only_pinned

    def validate_only_pinned(self):
        if (
            self._project.upload_state == constances.UploadState.EXTERNAL.value
            and self._only_pinned
        ):
            raise AppValidationException(
                f"Pin functionality is not supported for  projects containing {self._project.type} attached with URLs"
            )

    def validate_fuse(self):
        if (
            self._project.upload_state == constances.UploadState.EXTERNAL.value
            and self._include_fuse
        ):
            raise AppValidationException(
                f"Include fuse functionality is not supported for  projects containing {self._project.type} attached with URLs"
            )

    def execute(self):
        if self.is_valid():
            if self._project.upload_state == constances.UploadState.EXTERNAL.value:
                self._include_fuse = False

            if not self._annotation_statuses:
                self._annotation_statuses = (
                    constances.AnnotationStatus.IN_PROGRESS.name,
                    constances.AnnotationStatus.COMPLETED.name,
                    constances.AnnotationStatus.QUALITY_CHECK.name,
                    constances.AnnotationStatus.RETURNED.name,
                    constances.AnnotationStatus.NOT_STARTED.name,
                    constances.AnnotationStatus.SKIPPED.name,
                )

            response = self._backend_service.prepare_export(
                project_id=self._project.id,
                team_id=self._project.team_id,
                folders=self._folder_names,
                annotation_statuses=self._annotation_statuses,
                include_fuse=self._include_fuse,
                only_pinned=self._only_pinned,
            )
            if "error" in response:
                raise AppException(response["error"])

            report_message = ""
            if self._folder_names:
                report_message = f"[{', '.join(self._folder_names)}] "
            logger.info(
                f"Prepared export {response['name']} for project {self._project.name} "
                f"{report_message}(project ID {self._project.id})."
            )
            self._response.data = response

        return self._response


class GetExportsUseCase(BaseUseCase):
    def __init__(
        self,
        service: SuperannotateServiceProvider,
        project: ProjectEntity,
        return_metadata: bool = False,
    ):
        super().__init__()
        self._service = service
        self._project = project
        self._return_metadata = return_metadata

    def execute(self):
        if self.is_valid():
            data = self._service.get_exports(
                team_id=self._project.team_id, project_id=self._project.id
            )
            self._response.data = data
            if not self._return_metadata:
                self._response.data = [i["name"] for i in data]
        return self._response


class GetModelMetricsUseCase(BaseUseCase):
    def __init__(
        self,
        model_id: int,
        team_id: int,
        backend_service_provider: SuperannotateServiceProvider,
    ):
        super().__init__()
        self._model_id = model_id
        self._team_id = team_id
        self._backend_service = backend_service_provider

    def execute(self):
        metrics = self._backend_service.get_model_metrics(
            team_id=self._team_id, model_id=self._model_id
        )
        self._response.data = metrics
        return self._response


class DeleteMLModel(BaseUseCase):
    def __init__(self, model_id: int, models: BaseManageableRepository):
        super().__init__()
        self._model_id = model_id
        self._models = models

    def execute(self):
        self._response.data = self._models.delete(self._model_id)
        return self._response


class DownloadExportUseCase(BaseInteractiveUseCase):
    def __init__(
        self,
        service: SuperannotateServiceProvider,
        project: ProjectEntity,
        export_name: str,
        folder_path: str,
        extract_zip_contents: bool,
        to_s3_bucket: bool,
    ):
        super().__init__()
        self._service = service
        self._project = project
        self._export_name = export_name
        self._folder_path = folder_path
        self._extract_zip_contents = extract_zip_contents
        self._to_s3_bucket = to_s3_bucket
        self._temp_dir = None

    def upload_to_s3_from_folder(self, folder_path: str):
        to_s3_bucket = boto3.Session().resource("s3").Bucket(self._to_s3_bucket)
        files_to_upload = list(Path(folder_path).rglob("*.*"))

        def _upload_file_to_s3(_to_s3_bucket, _path, _s3_key) -> None:
            _to_s3_bucket.upload_file(_path, _s3_key)

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            results = []
            for path in files_to_upload:
                s3_key = f"{self._folder_path}/{path.name}"
                results.append(
                    executor.submit(_upload_file_to_s3, to_s3_bucket, str(path), s3_key)
                )
                yield

    def download_to_local_storage(self, destination: str):
        exports = self._service.get_exports(
            team_id=self._project.team_id, project_id=self._project.id
        )
        export = next(filter(lambda i: i["name"] == self._export_name, exports), None)
        export = self._service.get_export(
            team_id=self._project.team_id,
            project_id=self._project.id,
            export_id=export["id"],
        )
        if not export:
            raise AppException("Export not found.")
        export_status = export["status"]

        while export_status != ExportStatus.COMPLETE.value:
            logger.info("Waiting 5 seconds for export to finish on server.")
            time.sleep(5)

            export = self._service.get_export(
                team_id=self._project.team_id,
                project_id=self._project.id,
                export_id=export["id"],
            )
            if "error" in export:
                raise AppException(export["error"])
            export_status = export["status"]
            if export_status in (ExportStatus.ERROR.value, ExportStatus.CANCELED.value):
                raise AppException("Couldn't download export.")

        filename = Path(export["path"]).name
        filepath = Path(destination) / filename
        with requests.get(export["download"], stream=True) as response:
            response.raise_for_status()
            with open(filepath, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)
        if self._extract_zip_contents:
            with zipfile.ZipFile(filepath, "r") as f:
                f.extractall(destination)
            Path.unlink(filepath)
        return export["id"], filepath, destination

    def get_upload_files_count(self):
        if not self._temp_dir:
            self._temp_dir = tempfile.TemporaryDirectory()
            self.download_to_local_storage(self._temp_dir.name)
        return len(list(Path(self._temp_dir.name).rglob("*.*")))

    def execute(self):
        if self.is_valid():
            report = []
            if self._to_s3_bucket:
                self.get_upload_files_count()
                yield from self.upload_to_s3_from_folder(self._temp_dir.name)
                report.append(
                    f"Exported to AWS {self._to_s3_bucket}/{self._folder_path}"
                )
                self._temp_dir.cleanup()
            else:
                export_id, filepath, destination = self.download_to_local_storage(
                    self._folder_path
                )
                if self._extract_zip_contents:
                    report.append(f"Extracted {filepath} to folder {destination}")
                else:
                    report.append(f"Downloaded export ID {export_id} to {filepath}")
                yield
            self._response.data = "\n".join(report)
        return self._response


class DownloadMLModelUseCase(BaseUseCase):
    def __init__(
        self,
        model: MLModelEntity,
        download_path: str,
        backend_service_provider: SuperannotateServiceProvider,
        team_id: int,
    ):
        super().__init__()
        self._model = model
        self._download_path = download_path
        self._backend_service = backend_service_provider
        self._team_id = team_id

    def validate_training_status(self):
        if self._model.training_status not in [
            constances.TrainingStatus.COMPLETED.value,
            constances.TrainingStatus.FAILED_AFTER_EVALUATION_WITH_SAVE_MODEL.value,
        ]:
            raise AppException("Unable to download.")

    def execute(self):
        if self.is_valid():
            metrics_name = os.path.basename(self._model.path).replace(".pth", ".json")
            mapper_path = self._model.config_path.replace(
                os.path.basename(self._model.config_path), "classes_mapper.json"
            )
            metrics_path = self._model.config_path.replace(
                os.path.basename(self._model.config_path), metrics_name
            )

            auth_response = self._backend_service.get_ml_model_download_tokens(
                self._team_id, self._model.uuid
            )
            if not auth_response.ok:
                raise AppException(auth_response.error)
            s3_session = boto3.Session(
                aws_access_key_id=auth_response.data.access_key,
                aws_secret_access_key=auth_response.data.secret_key,
                aws_session_token=auth_response.data.session_token,
                region_name=auth_response.data.region,
            )
            bucket = s3_session.resource("s3").Bucket(auth_response.data.bucket)

            bucket.download_file(
                self._model.config_path,
                os.path.join(self._download_path, "config.yaml"),
            )
            bucket.download_file(
                self._model.path,
                os.path.join(self._download_path, os.path.basename(self._model.path)),
            )
            try:
                bucket.download_file(
                    metrics_path, os.path.join(self._download_path, metrics_name)
                )
                bucket.download_file(
                    mapper_path,
                    os.path.join(self._download_path, "classes_mapper.json"),
                )
            except ClientError:
                logger.info(
                    "The specified model does not contain a classes_mapper and/or a metrics file."
                )
            self._response.data = self._model
        return self._response


class BenchmarkUseCase(BaseUseCase):
    def __init__(
        self,
        project: ProjectEntity,
        ground_truth_folder_name: str,
        folder_names: list,
        export_dir: str,
        image_list: list,
        annotation_type: str,
        show_plots: bool,
    ):
        super().__init__()
        self._project = project
        self._ground_truth_folder_name = ground_truth_folder_name
        self._folder_names = folder_names
        self._export_dir = export_dir
        self._image_list = image_list
        self._annotation_type = annotation_type
        self._show_plots = show_plots

    def execute(self):
        project_df = aggregate_image_annotations_as_df(self._export_dir)
        gt_project_df = project_df[
            project_df["folderName"] == self._ground_truth_folder_name
        ]
        benchmark_dfs = []
        for folder_name in self._folder_names:
            folder_df = project_df[project_df["folderName"] == folder_name]
            project_gt_df = pd.concat([folder_df, gt_project_df])
            project_gt_df = project_gt_df[project_gt_df["instanceId"].notna()]

            if self._image_list is not None:
                project_gt_df = project_gt_df.loc[
                    project_gt_df["imageName"].isin(self._image_list)
                ]

            project_gt_df.query("type == '" + self._annotation_type + "'", inplace=True)

            project_gt_df = project_gt_df.groupby(
                ["imageName", "instanceId", "folderName"]
            )

            def aggregate_attributes(instance_df):
                def attribute_to_list(attribute_df):
                    attribute_names = list(attribute_df["attributeName"])
                    attribute_df["attributeNames"] = len(attribute_df) * [
                        attribute_names
                    ]
                    return attribute_df

                attributes = None
                if not instance_df["attributeGroupName"].isna().all():
                    attrib_group_name = instance_df.groupby("attributeGroupName")[
                        ["attributeGroupName", "attributeName"]
                    ].apply(attribute_to_list)
                    attributes = dict(
                        zip(
                            attrib_group_name["attributeGroupName"],
                            attrib_group_name["attributeNames"],
                        )
                    )

                instance_df.drop(
                    ["attributeGroupName", "attributeName"], axis=1, inplace=True
                )
                instance_df.drop_duplicates(
                    subset=["imageName", "instanceId", "folderName"], inplace=True
                )
                instance_df["attributes"] = [attributes]
                return instance_df

            project_gt_df = project_gt_df.apply(aggregate_attributes).reset_index(
                drop=True
            )
            unique_images = set(project_gt_df["imageName"])
            all_benchmark_data = []
            for image_name in unique_images:
                image_data = image_consensus(
                    project_gt_df, image_name, self._annotation_type
                )
                all_benchmark_data.append(pd.DataFrame(image_data))
            benchmark_project_df = pd.concat(all_benchmark_data, ignore_index=True)
            benchmark_project_df = benchmark_project_df[
                benchmark_project_df["folderName"] == folder_name
            ]
            benchmark_dfs.append(benchmark_project_df)
        benchmark_df = pd.concat(benchmark_dfs, ignore_index=True)
        if self._show_plots:
            consensus_plot(benchmark_df, self._folder_names)
        self._response.data = benchmark_df
        return self._response


class ConsensusUseCase(BaseUseCase):
    def __init__(
        self,
        project: ProjectEntity,
        folder_names: list,
        export_dir: str,
        image_list: list,
        annotation_type: str,
        show_plots: bool,
    ):
        super().__init__()
        self._project = project
        self._folder_names = folder_names
        self._export_dir = export_dir
        self._image_list = image_list
        self._annota_type_type = annotation_type
        self._show_plots = show_plots

    def execute(self):
        project_df = aggregate_image_annotations_as_df(self._export_dir)
        all_projects_df = project_df[project_df["instanceId"].notna()]
        all_projects_df = all_projects_df.loc[
            all_projects_df["folderName"].isin(self._folder_names)
        ]

        if self._image_list is not None:
            all_projects_df = all_projects_df.loc[
                all_projects_df["imageName"].isin(self._image_list)
            ]

        all_projects_df.query("type == '" + self._annota_type_type + "'", inplace=True)

        def aggregate_attributes(instance_df):
            def attribute_to_list(attribute_df):
                attribute_names = list(attribute_df["attributeName"])
                attribute_df["attributeNames"] = len(attribute_df) * [attribute_names]
                return attribute_df

            attributes = None
            if not instance_df["attributeGroupName"].isna().all():
                attrib_group_name = instance_df.groupby("attributeGroupName")[
                    ["attributeGroupName", "attributeName"]
                ].apply(attribute_to_list)
                attributes = dict(
                    zip(
                        attrib_group_name["attributeGroupName"],
                        attrib_group_name["attributeNames"],
                    )
                )

            instance_df.drop(
                ["attributeGroupName", "attributeName"], axis=1, inplace=True
            )
            instance_df.drop_duplicates(
                subset=["imageName", "instanceId", "folderName"], inplace=True
            )
            instance_df["attributes"] = [attributes]
            return instance_df

        all_projects_df = all_projects_df.groupby(
            ["imageName", "instanceId", "folderName"]
        )
        all_projects_df = all_projects_df.apply(aggregate_attributes).reset_index(
            drop=True
        )
        unique_images = set(all_projects_df["imageName"])
        all_consensus_data = []
        for image_name in unique_images:
            image_data = image_consensus(
                all_projects_df, image_name, self._annota_type_type
            )
            all_consensus_data.append(pd.DataFrame(image_data))

        consensus_df = pd.concat(all_consensus_data, ignore_index=True)

        if self._show_plots:
            consensus_plot(consensus_df, self._folder_names)

        self._response.data = consensus_df
        return self._response


class RunPredictionUseCase(BaseUseCase):
    def __init__(
        self,
        project: ProjectEntity,
        ml_model_repo: BaseManageableRepository,
        ml_model_name: str,
        images_list: list,
        service: SuperannotateServiceProvider,
        folder: FolderEntity,
    ):
        super().__init__()
        self._project = project
        self._ml_model_repo = ml_model_repo
        self._ml_model_name = ml_model_name
        self._images_list = images_list
        self._service = service
        self._folder = folder

    def validate_project_type(self):
        if self._project.type in constances.LIMITED_FUNCTIONS:
            raise AppValidationException(
                constances.LIMITED_FUNCTIONS[self._project.type]
            )

    def execute(self):
        if self.is_valid():
            images = (
                GetBulkImages(
                    service=self._service,
                    project_id=self._project.id,
                    team_id=self._project.team_id,
                    folder_id=self._folder.uuid,
                    images=self._images_list,
                )
                .execute()
                .data
            )

            image_ids = [image.uuid for image in images]
            image_names = [image.name for image in images]

            if not len(image_names):
                self._response.errors = AppException(
                    "No valid image names were provided."
                )
                return self._response

            ml_models = self._ml_model_repo.get_all(
                condition=Condition("name", self._ml_model_name, EQ)
                & Condition("include_global", True, EQ)
                & Condition("team_id", self._project.team_id, EQ)
            )
            ml_model = None
            for model in ml_models:
                if model.name == self._ml_model_name:
                    ml_model = model

            res = self._service.run_prediction(
                team_id=self._project.team_id,
                project_id=self._project.id,
                ml_model_id=ml_model.uuid,
                image_ids=image_ids,
            )
            if not res.ok:
                return self._response

            success_images = []
            failed_images = []
            while len(success_images) + len(failed_images) != len(image_ids):
                images_metadata = (
                    GetBulkImages(
                        service=self._service,
                        project_id=self._project.id,
                        team_id=self._project.team_id,
                        folder_id=self._folder.uuid,
                        images=self._images_list,
                    )
                    .execute()
                    .data
                )

                success_images = [
                    img.name
                    for img in images_metadata
                    if img.prediction_status
                    == constances.SegmentationStatus.COMPLETED.value
                ]
                failed_images = [
                    img.name
                    for img in images_metadata
                    if img.prediction_status
                    == constances.SegmentationStatus.FAILED.value
                ]

                complete_images = success_images + failed_images
                logger.info(
                    f"prediction complete on {len(complete_images)} / {len(image_ids)} images"
                )
                time.sleep(5)

            self._response.data = (success_images, failed_images)
        return self._response


class SearchMLModels(BaseUseCase):
    def __init__(
        self, ml_models_repo: BaseManageableRepository, condition: Condition,
    ):
        super().__init__()
        self._ml_models = ml_models_repo
        self._condition = condition

    def execute(self):
        ml_models = self._ml_models.get_all(condition=self._condition)
        ml_models = [ml_model.to_dict() for ml_model in ml_models]
        self._response.data = ml_models
        return self._response



from pathlib import Path

import numpy as np

from ....common import write_to_json
from ..baseStrategy import baseStrategy


class VoTTStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        json_data = self.get_file_list()
        classes = self.conversion_algorithm(json_data, self.task, self.output_dir)
        sa_classes = self._create_classes(classes)
        (self.output_dir / "classes").mkdir(exist_ok=True)
        write_to_json(self.output_dir / "classes" / "classes.json", sa_classes)

    def get_file_list(self):
        json_file_list = []
        path = Path(self.export_root)
        if self.dataset_name != "":
            json_file_list.append(path.joinpath(self.dataset_name + ".json"))
        else:
            file_generator = path.glob("*.json")
            for gen in file_generator:
                json_file_list.append(gen)

        return json_file_list

    def _create_classes(self, classes_map):
        classes_loader = []
        for key in classes_map:
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            sa_classes = {"name": key, "color": hexcolor, "attribute_groups": []}
            classes_loader.append(sa_classes)
        return classes_loader



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Extra


class ErrorMessage(BaseModel):
    error: str


class Limit(BaseModel):
    max_image_count: Optional[int]
    remaining_image_count: int


class UserLimits(BaseModel):
    user_limit: Optional[Limit]
    project_limit: Limit
    folder_limit: Limit


class UploadAnnotationAuthData(BaseModel):
    access_key: str
    secret_key: str
    session_token: str
    region: str
    bucket: str
    images: Dict[int, dict]

    class Config:
        extra = Extra.allow
        fields = {
            "access_key": "accessKeyId",
            "secret_key": "secretAccessKey",
            "session_token": "sessionToken",
            "region": "region",
        }

    def __init__(self, **data):
        credentials = data["creds"]
        data.update(credentials)
        del data["creds"]
        super().__init__(**data)


class DownloadMLModelAuthData(BaseModel):
    access_key: str
    secret_key: str
    session_token: str
    region: str
    bucket: str
    paths: List[str]

    class Config:
        extra = Extra.allow
        fields = {
            "access_key": "accessKeyId",
            "secret_key": "secretAccessKey",
            "session_token": "sessionToken",
            "region": "region",
        }

    def __init__(self, **data):
        credentials = data["tokens"]
        data.update(credentials)
        del data["tokens"]
        super().__init__(**data)


class ServiceResponse(BaseModel):
    status: int
    reason: str
    content: Union[bytes, str]
    data: Any

    def __init__(self, response, content_type=None):
        data = {
            "status": response.status_code,
            "reason": response.reason,
            "content": response.content,
        }
        if response.ok:
            if content_type:
                data["data"] = content_type(**response.json())
            else:
                data["data"] = response.json()
        super().__init__(**data)

    @property
    def ok(self):
        return 199 < self.status < 300

    @property
    def error(self):
        return getattr(self.data, "error", "Unknown error.")



import os
import sys


sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


def get_default_controller():
    from lib.infrastructure.controller import Controller

    return Controller.get_default()



import uuid
from ast import literal_eval
from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import boto3
import pandas as pd
from superannotate.lib.app.exceptions import AppException
from superannotate.lib.app.exceptions import PathError
from superannotate.lib.core import ATTACHED_VIDEO_ANNOTATION_POSTFIX
from superannotate.lib.core import PIXEL_ANNOTATION_POSTFIX
from superannotate.lib.core import VECTOR_ANNOTATION_POSTFIX


def split_project_path(project_path: str) -> Tuple[str, Optional[str]]:
    path = Path(project_path)
    if len(path.parts) > 3:
        raise PathError("There can be no sub folders in the project")
    elif len(path.parts) == 2:
        project_name, folder_name = path.parts
    else:
        project_name, folder_name = path.name, ""

    return project_name, folder_name


def extract_project_folder(user_input: Union[str, dict]) -> Tuple[str, Optional[str]]:
    if isinstance(user_input, str):
        return split_project_path(user_input)
    elif isinstance(user_input, dict):
        project_path = user_input.get("name")
        if not project_path:
            raise PathError("Invalid project path")
        return split_project_path(user_input["name"])
    raise PathError("Invalid project path")


def get_annotation_paths(folder_path, s3_bucket=None, recursive=False):
    annotation_paths = []
    if s3_bucket:
        return get_s3_annotation_paths(
            folder_path, s3_bucket, annotation_paths, recursive
        )
    return get_local_annotation_paths(folder_path, set(annotation_paths), recursive)


def get_local_annotation_paths(
    folder_path: Union[str, Path], annotation_paths: set, recursive: bool
) -> List[str]:
    all_items = [*Path(folder_path).glob("*")]
    all_folders = [i for i in all_items if i.is_dir()]
    all_not_folders = [i for i in all_items if not i.is_dir()]
    annotation_paths.update(
        [
            str(i)
            for i in all_not_folders
            if i.name.endswith(
                (
                    VECTOR_ANNOTATION_POSTFIX,
                    PIXEL_ANNOTATION_POSTFIX,
                    ATTACHED_VIDEO_ANNOTATION_POSTFIX,
                )
            )
        ]
    )
    if recursive:
        for folder in all_folders:
            get_local_annotation_paths(
                folder_path=folder,
                annotation_paths=annotation_paths,
                recursive=recursive,
            )
    return list(annotation_paths)


def get_s3_annotation_paths(folder_path, s3_bucket, annotation_paths, recursive):
    s3_client = boto3.client("s3")
    result = s3_client.list_objects(Bucket=s3_bucket, Prefix=folder_path, Delimiter="/")
    result = result.get("CommonPrefixes")
    if recursive and result:
        for folder in result:
            get_s3_annotation_paths(
                folder.get("Prefix"), s3_bucket, annotation_paths, recursive
            )

    paginator = s3_client.get_paginator("list_objects_v2")
    for data in paginator.paginate(Bucket=s3_bucket, Prefix=folder_path):
        for annotation in data.get("Contents", []):
            key = annotation["Key"]
            if (
                key.endswith(VECTOR_ANNOTATION_POSTFIX)
                or key.endswith(PIXEL_ANNOTATION_POSTFIX)
                or key.endswith(ATTACHED_VIDEO_ANNOTATION_POSTFIX)
            ):
                if not recursive and "/" in key[len(folder_path) + 1 :]:
                    continue
                annotation_paths.append(key)
    return list(set(annotation_paths))


def reformat_metrics_json(data, name):
    continuous_metrics = []
    per_evaluation_metrics = []
    for item in data:
        for key in item:
            try:
                item[key] = literal_eval(item[key])
            except Exception as e:
                pass
        if "data_time" in item and item["eta_seconds"] != 0:
            continuous_metrics += [item]
        else:
            per_evaluation_metrics += [item]
    continuous_metrics_df = pd.DataFrame.from_dict(continuous_metrics)
    per_evaluation_metrics_df = pd.DataFrame.from_dict(per_evaluation_metrics)
    continuous_metrics_df = drop_non_plotable_cols(continuous_metrics_df)
    per_evaluation_metrics_df = drop_non_plotable_cols(per_evaluation_metrics_df)
    continuous_metrics_df["model"] = name
    per_evaluation_metrics_df["model"] = name
    if "total_loss" in per_evaluation_metrics_df:
        per_evaluation_metrics_df = per_evaluation_metrics_df.drop(columns="total_loss")

    per_evaluation_metrics_df = per_evaluation_metrics_df.dropna(axis="rows")
    return continuous_metrics_df, per_evaluation_metrics_df


def drop_non_plotable_cols(df):
    for column in df:
        if metric_is_plottable(column):
            continue
        df = df.drop(columns=column)
    return df


def metric_is_plottable(key):
    if key == "total_loss" or "mIoU" in key or "mAP" in key or key == "iteration":
        return True
    return False


def get_paths_and_duplicated_from_csv(csv_path):
    image_data = pd.read_csv(csv_path, dtype=str)
    image_data = image_data[~image_data["url"].isnull()]
    if "name" in image_data.columns:
        image_data["name"] = (
            image_data["name"]
            .fillna("")
            .apply(lambda cell: cell if str(cell).strip() else str(uuid.uuid4()))
        )
    else:
        image_data["name"] = [str(uuid.uuid4()) for _ in range(len(image_data.index))]

    image_data = pd.DataFrame(image_data, columns=["name", "url"])
    img_names_urls = image_data.rename(columns={"url": "path"}).to_dict(
        orient="records"
    )
    duplicate_images = []
    seen = []
    images_to_upload = []
    for i in img_names_urls:
        temp = i["name"]
        i["name"] = i["name"].strip()
        if i["name"] not in seen:
            seen.append(i["name"])
            images_to_upload.append(i)
        else:
            duplicate_images.append(temp)
    return images_to_upload, duplicate_images


def get_name_url_duplicated_from_csv(csv_path):
    image_data = pd.read_csv(csv_path, dtype=str)
    if "url" not in image_data.columns:
        raise AppException("Column 'url' is required")
    image_data = image_data[~image_data["url"].isnull()]
    if "name" in image_data.columns:
        image_data["name"] = (
            image_data["name"]
            .fillna("")
            .apply(lambda cell: cell if str(cell).strip() else str(uuid.uuid4()))
        )
    else:
        image_data["name"] = [str(uuid.uuid4()) for _ in range(len(image_data.index))]

    image_data = pd.DataFrame(image_data, columns=["name", "url"])
    img_names_urls = image_data.to_dict(orient="records")
    duplicate_images = []
    seen = []
    images_to_upload = []
    for i in img_names_urls:
        temp = i["name"]
        i["name"] = i["name"].strip()
        if i["name"] not in seen:
            seen.append(i["name"])
            images_to_upload.append(i)
        else:
            duplicate_images.append(temp)
    return images_to_upload, duplicate_images



"""
Labelbox to SA conversion method
"""
import threading
from pathlib import Path

import cv2
import numpy as np
from superannotate.logger import get_default_logger

from ....common import blue_color_generator
from ....common import hex_to_rgb
from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_pixel_instance
from ..sa_json_helper import _create_sa_json
from .labelbox_helper import _create_attributes_list
from .labelbox_helper import _create_classes_id_map
from .labelbox_helper import image_downloader

logger = get_default_logger()


def labelbox_instance_segmentation_to_sa_pixel(json_data, output_dir, input_dir):
    classes = _create_classes_id_map(json_data)

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(json_data), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for data in json_data:
        file_name = data["External ID"] + "___pixel.json"
        mask_name = data["External ID"] + "___save.png"
        sa_metadata = {"name": data["External ID"]}
        if "objects" not in data["Label"].keys():
            sa_json = _create_sa_json([], sa_metadata)
            write_to_json(output_dir / file_name, sa_json)
            continue

        instances = data["Label"]["objects"]
        sa_instances = []
        blue_colors = blue_color_generator(len(instances))

        for i, instance in enumerate(instances):
            class_name = instance["value"]
            attributes = []
            if "classifications" in instance.keys():
                attributes = _create_attributes_list(instance["classifications"])

            if (
                "bbox" in instance.keys()
                or "polygon" in instance.keys()
                or "line" in instance.keys()
                or "point" in instance.keys()
            ):
                continue

            bitmask_name = "%s.png" % instance["featureId"]
            downloaded = image_downloader(instance["instanceURI"], bitmask_name)
            if downloaded:
                mask = cv2.imread(bitmask_name)
            else:
                mask = cv2.imread(str(input_dir / "bitmasks" / bitmask_name))
                bitmask_name = output_dir / bitmask_name

            if isinstance(mask, type(None)):
                logger.warning("Can't open '%s' bitmask.", bitmask_name)
                images_not_converted.append(bitmask_name)
                continue

            if i == 0:
                H, W, _ = mask.shape
                sa_metadata["width"] = W
                sa_metadata["height"] = H
                sa_mask = np.zeros((H, W, 4))
            sa_mask[np.all(mask == [255, 255, 255], axis=2)] = list(
                hex_to_rgb(blue_colors[i])
            )[::-1] + [255]

            parts = [{"color": blue_colors[i]}]
            sa_obj = _create_pixel_instance(parts, attributes, class_name)

            sa_instances.append(sa_obj.copy())
            Path(bitmask_name).unlink()

        images_converted.append(data["External ID"])
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
        cv2.imwrite(str(output_dir / mask_name), sa_mask)

    finish_event.set()
    tqdm_thread.join()
    return classes



import os
import xml.etree.ElementTree as ET


def _get_voc_instances_from_xml(file_path):
    with open(os.path.splitext(file_path)[0] + ".xml") as f:
        tree = ET.parse(f)
    instances = tree.findall("object")
    voc_instances = []
    for instance in instances:
        class_name = instance.find("name").text
        class_attributes = []
        for attr in ["pose", "occluded", "difficult", "truncated"]:
            attr_value = instance.find(attr)
            if attr_value is not None:
                class_attributes.append({"name": attr_value.text, "groupName": attr})

        bbox = instance.find("bndbox")
        bbox = [float(bbox.find(x).text) for x in ["xmin", "ymin", "xmax", "ymax"]]
        voc_instances.append(({class_name: class_attributes}, bbox))
    return voc_instances


def _iou(bbox1, bbox2):
    xmin1, ymin1, xmax1, ymax1 = bbox1
    xmin2, ymin2, xmax2, ymax2 = bbox2

    x = max(0, min(xmax1, xmax2) - max(xmin1, xmin2))
    y = max(0, min(ymax1, ymax2) - max(ymin1, ymin2))
    return (
        x
        * y
        / float(
            (xmax1 - xmin1) * (ymax1 - ymin1)
            + (xmax2 - xmin2) * (ymax2 - ymin2)
            - x * y
        )
    )


def _get_image_shape_from_xml(file_path):
    with open(os.path.splitext(file_path)[0] + ".xml") as f:
        tree = ET.parse(f)

    size = tree.find("size")
    width = int(size.find("width").text)
    height = int(size.find("height").text)

    return height, width



from typing import List

import lib.core as constances
from lib.core.conditions import Condition
from lib.core.conditions import CONDITION_EQ as EQ
from lib.core.entities import FolderEntity
from lib.core.entities import ProjectEntity
from lib.core.exceptions import AppException
from lib.core.exceptions import AppValidationException
from lib.core.repositories import BaseManageableRepository
from lib.core.repositories import BaseReadOnlyRepository
from lib.core.serviceproviders import SuperannotateServiceProvider
from lib.core.usecases.base import BaseUseCase
from superannotate.logger import get_default_logger

logger = get_default_logger()


class CreateFolderUseCase(BaseUseCase):
    def __init__(
        self,
        project: ProjectEntity,
        folder: FolderEntity,
        folders: BaseManageableRepository,
    ):
        super().__init__()
        self._project = project
        self._folder = folder
        self._folders = folders
        self._origin_name = folder.name

    def validate_folder(self):
        if not self._folder.name:
            raise AppValidationException("Folder name cannot be empty.")
        if (
            len(
                set(self._folder.name).intersection(
                    constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                )
            )
            > 0
        ):
            self._folder.name = "".join(
                "_"
                if char in constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                else char
                for char in self._folder.name
            )
            logger.warning(
                "New folder name has special characters. Special characters will be replaced by underscores."
            )

    def execute(self):
        if self.is_valid():
            self._folder.project_id = self._project.id
            self._response.data = self._folders.insert(self._folder)
            if self._response.data.name not in (self._origin_name, self._folder.name):
                logger.warning(
                    f"Created folder has name {self._response.data.name},"
                    f" since folder with name {self._folder.name} already existed."
                )
        return self._response


class GetFolderUseCase(BaseUseCase):
    def __init__(
        self,
        project: ProjectEntity,
        folders: BaseReadOnlyRepository,
        folder_name: str,
        team_id: int,
    ):
        super().__init__()
        self._project = project
        self._folders = folders
        self._folder_name = folder_name
        self._team_id = team_id

    def execute(self):
        condition = (
            Condition("name", self._folder_name, EQ)
            & Condition("team_id", self._team_id, EQ)
            & Condition("project_id", self._project.id, EQ)
        )
        try:
            self._response.data = self._folders.get_one(condition)
        except AppException as e:
            self._response.errors = e
        return self._response


class SearchFoldersUseCase(BaseUseCase):
    def __init__(
        self,
        project: ProjectEntity,
        folders: BaseReadOnlyRepository,
        condition: Condition,
        folder_name: str = None,
        include_users=False,
    ):
        super().__init__()
        self._project = project
        self._folders = folders
        self._folder_name = folder_name
        self._condition = condition
        self._include_users = include_users

    def execute(self):
        condition = (
            self._condition
            & Condition("project_id", self._project.id, EQ)
            & Condition("team_id", self._project.team_id, EQ)
            & Condition("includeUsers", self._include_users, EQ)
        )
        if self._folder_name:
            condition &= Condition("name", self._folder_name, EQ)
        self._response.data = self._folders.get_all(condition)
        return self._response


class DeleteFolderUseCase(BaseUseCase):
    def __init__(
        self,
        project: ProjectEntity,
        folders: BaseManageableRepository,
        folders_to_delete: List[FolderEntity],
    ):
        super().__init__()
        self._project = project
        self._folders = folders
        self._folders_to_delete = folders_to_delete

    def execute(self):
        if self._folders_to_delete:
            is_deleted = self._folders.bulk_delete(self._folders_to_delete)
            if not is_deleted:
                self._response.errors = AppException("Couldn't delete folders.")
        else:
            self._response.errors = AppException("There is no folder to delete.")
        return self._response


class UpdateFolderUseCase(BaseUseCase):
    def __init__(
        self, folders: BaseManageableRepository, folder: FolderEntity,
    ):
        super().__init__()
        self._folders = folders
        self._folder = folder

    def validate_folder(self):
        if not self._folder.name:
            raise AppValidationException("Folder name cannot be empty.")
        if (
            len(
                set(self._folder.name).intersection(
                    constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                )
            )
            > 0
        ):
            self._folder.name = "".join(
                "_"
                if char in constances.SPECIAL_CHARACTERS_IN_PROJECT_FOLDER_NAMES
                else char
                for char in self._folder.name
            )
            logger.warning(
                "New folder name has special characters. Special characters will be replaced by underscores."
            )

    def execute(self):
        if self.is_valid():
            folder = self._folders.update(self._folder)
            if not folder:
                self._response.errors = AppException("Couldn't rename folder.")
            self._response.data = folder
        return self._response


class AssignFolderUseCase(BaseUseCase):
    def __init__(
        self,
        service: SuperannotateServiceProvider,
        project_entity: ProjectEntity,
        folder: FolderEntity,
        users: List[str],
    ):
        super().__init__()
        self._service = service
        self._project_entity = project_entity
        self._folder = folder
        self._users = users

    def execute(self):
        is_assigned = self._service.assign_folder(
            team_id=self._project_entity.team_id,
            project_id=self._project_entity.id,
            folder_name=self._folder.name,
            users=self._users,
        )
        if is_assigned:
            logger.info(
                f'Assigned {self._folder.name} to users: {", ".join(self._users)}'
            )
        else:
            self._response.errors = AppException(
                f"Couldn't assign folder to users: {', '.join(self._users)}"
            )
        return self._response



from typing import List

from lib.core.entities import FolderEntity
from lib.core.entities import IntegrationEntity
from lib.core.entities import ProjectEntity
from lib.core.entities import TeamEntity
from lib.core.exceptions import AppException
from lib.core.reporter import Reporter
from lib.core.repositories import BaseReadOnlyRepository
from lib.core.response import Response
from lib.core.serviceproviders import SuperannotateServiceProvider
from lib.core.usecases import BaseReportableUseCae


class GetIntegrations(BaseReportableUseCae):
    def __init__(
        self,
        reporter: Reporter,
        team: TeamEntity,
        integrations: BaseReadOnlyRepository,
    ):

        super().__init__(reporter)
        self._team = team
        self._integrations = integrations

    def execute(self) -> Response:
        integrations = self._integrations.get_all()
        integrations = list(sorted(integrations, key=lambda x: x.createdAt))
        integrations.reverse()
        self._response.data = integrations
        return self._response


class AttachIntegrations(BaseReportableUseCae):
    def __init__(
        self,
        reporter: Reporter,
        team: TeamEntity,
        project: ProjectEntity,
        folder: FolderEntity,
        integrations: BaseReadOnlyRepository,
        backend_service: SuperannotateServiceProvider,
        integration: IntegrationEntity,
        folder_path: str = None,
    ):

        super().__init__(reporter)
        self._project = project
        self._folder = folder
        self._team = team
        self._integration = integration
        self._client = backend_service
        self._folder_path = folder_path
        self._integrations = integrations

    @property
    def _upload_path(self):
        return f"{self._project.name}{f'/{self._folder.name}' if self._folder.name != 'root' else ''}"

    def execute(self) -> Response:
        integrations: List[IntegrationEntity] = self._integrations.get_all()
        integration_name_lower = self._integration.name.lower()
        integration = next(
            (i for i in integrations if i.name.lower() == integration_name_lower), None
        )
        if integration:
            self.reporter.log_info(
                "Attaching file(s) from "
                f"{integration.root}{f'/{self._folder_path}' if self._folder_path else ''} "
                f"to {self._upload_path}. This may take some time."
            )
            attached = self._client.attach_integrations(
                self._team.uuid,
                self._project.id,
                integration.id,
                self._folder.uuid,
                self._folder_path,
            )
            if not attached:
                self._response.errors = AppException(
                    f"An error occurred for {self._integration.name}. Please make sure: "
                    "\n - The bucket exists."
                    "\n - The connection is valid."
                    "\n - The path to a specified directory is correct."
                )
        else:
            self._response.errors = AppException("Integration not found.")
        return self._response



def _update_classes_dict(classes, new_class, new_attributes):
    if new_class not in classes.keys():
        classes[new_class] = {"attribute_group": {}}
        classes[new_class]["attribute_group"]["converted_attributes"] = {}
        classes[new_class]["attribute_group"]["converted_attributes"][
            "is_multiselect"
        ] = 1
        classes[new_class]["attribute_group"]["converted_attributes"][
            "attributes"
        ] = new_attributes
    else:
        classes[new_class]["attribute_group"]["converted_attributes"][
            "attributes"
        ] += new_attributes
    return classes


def _create_attributes_list(dl_attributes):
    attributes = []
    for attribute in dl_attributes:
        attr = {"name": attribute, "groupName": "converted_attributes"}
        attributes.append(attr)
    return attributes
